from __future__ import annotations

import logging
from enum import Enum
from operator import add
from typing import TYPE_CHECKING, Callable, Dict, List, Literal, Optional, Union

import numpy as np
import pandas as pd
from pint_pandas import PintType
from pint_pandas.pint_array import PintSeriesAccessor
from pydantic import (
    BaseModel,
    ConfigDict,
    GetJsonSchemaHandler,
    ValidationError,
    field_validator,
    model_validator,
)
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import CoreSchema

import ITR

from .configs import LoggingConfig, ProjectionControls
from .data.osc_units import (
    PA_,
    Q_,
    BenchmarkMetric,
    BenchmarkQuantity,
    EI_Metric,
    EI_Quantity,
    EmissionsMetric,
    EmissionsQuantity,
    MonetaryQuantity,
    ProductionMetric,
    ProductionQuantity,
    Quantity_type,
    ureg,
)

logger = logging.getLogger(__name__)
LoggingConfig.add_config_to_logger(logger)


class SortableEnum(Enum):
    def __str__(self):
        return self.name

    def __ge__(self, other):
        if self.__class__ is other.__class__:
            order = list(self.__class__)
            return order.index(self) >= order.index(other)
        return NotImplemented

    def __gt__(self, other):
        if self.__class__ is other.__class__:
            order = list(self.__class__)
            return order.index(self) > order.index(other)
        return NotImplemented

    def __le__(self, other):
        if self.__class__ is other.__class__:
            order = list(self.__class__)
            return order.index(self) <= order.index(other)
        return NotImplemented

    def __lt__(self, other):
        if self.__class__ is other.__class__:
            order = list(self.__class__)
            return order.index(self) < order.index(other)
        return NotImplemented


class EScope(SortableEnum):
    S1 = "S1"
    S2 = "S2"
    S3 = "S3"
    S1S2 = "S1+S2"
    S1S2S3 = "S1+S2+S3"
    AnyScope = "AnyScope"

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.__str__()

    @classmethod
    def get_scopes(cls) -> List[str]:
        """
        Get a list of all scopes.
        :return: A list of EScope string values
        """
        return ["S1", "S2", "S3", "S1S2", "S1S2S3"]

    @classmethod
    def get_result_scopes(cls) -> List["EScope"]:
        """
        Get a list of scopes that should be calculated if the user leaves it open.

        :return: A list of EScope objects
        """
        # FIXME: Should this also contain cls.S2 or no?
        return [cls.S1, cls.S1S2, cls.S3, cls.S1S2S3]


class ETimeFrames(SortableEnum):
    """
    TODO: add support for multiple timeframes. Long currently corresponds to 2050.
    """

    SHORT = "short"
    MID = "mid"
    LONG = "long"


class ECarbonBudgetScenario(Enum):
    P25 = "25 percentile"
    P75 = "75 percentile"
    MEAN = "Average"


class EScoreResultType(Enum):
    DEFAULT = "Default"
    TRAJECTORY_ONLY = "Trajectory only"
    TARGET_ONLY = "Target only"
    COMPLETE = "Complete"

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.__str__()

    @classmethod
    def get_result_types(cls) -> List[str]:
        """
        Get a list of all result types, ordered by priority (first << last priority).
        :return: A list of the EScoreResultType values
        """
        return [
            EScoreResultType.DEFAULT,
            EScoreResultType.TRAJECTORY_ONLY,
            EScoreResultType.TARGET_ONLY,
            EScoreResultType.COMPLETE,
        ]


class AggregationContribution(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    company_name: str
    company_id: str
    temperature_score: Quantity_type("delta_degC")
    contribution_relative: Optional[Quantity_type("percent")] = None
    contribution: Optional[Quantity_type("delta_degC")] = None

    def __getitem__(self, item):
        return getattr(self, item)


class Aggregation(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    score: Quantity_type("delta_degC")
    # proportion is a number from 0..1
    proportion: float
    contributions: List[AggregationContribution]

    def __getitem__(self, item):
        return getattr(self, item)


class ScoreAggregation(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    all: Aggregation
    influence_percentage: Quantity_type("percent")
    grouped: Dict[str, Aggregation]

    def __getitem__(self, item):
        return getattr(self, item)


class ScoreAggregationScopes(BaseModel):
    S1: Optional[ScoreAggregation] = None
    S2: Optional[ScoreAggregation] = None
    S1S2: Optional[ScoreAggregation] = None
    S3: Optional[ScoreAggregation] = None
    S1S2S3: Optional[ScoreAggregation] = None

    def __getitem__(self, item):
        return getattr(self, item)


class ScoreAggregations(BaseModel):
    short: Optional[ScoreAggregationScopes] = None
    mid: Optional[ScoreAggregationScopes] = None
    long: Optional[ScoreAggregationScopes] = None

    def __getitem__(self, item):
        return getattr(self, item)


class PortfolioCompany(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    company_name: str
    company_id: str
    company_isin: Optional[str] = None
    investment_value: MonetaryQuantity
    user_fields: Optional[dict] = None


# U is Unquantified, which is presently how our benchmarks come in (production_metric comes in elsewhere)
class UProjection(BaseModel):
    year: int
    value: float


class IProjection(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    year: int
    value: BenchmarkQuantity


class IBenchmark(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    sector: str
    region: str
    benchmark_metric: BenchmarkMetric
    projections_nounits: Optional[List[UProjection]] = None
    projections: Optional[List[IProjection]] = None
    base_year_production: Optional[ProductionQuantity] = None  # FIXME: applies only to production benchmarks

    def __init__(
        self,
        benchmark_metric,
        projections_nounits=None,
        projections=None,
        base_year_production=None,
        *args,
        **kwargs,
    ):
        # FIXME: Probably want to define `target_end_year` to be 2051, not 2050...
        super().__init__(
            benchmark_metric=benchmark_metric,
            projections_nounits=projections_nounits,
            projections=projections,
            base_year_production=base_year_production,
            *args,
            **kwargs,
        )
        # Sadly we need to build the full projection range before cutting it down to size...
        # ...until Tiemann learns the bi-valence of dict and Model parameters
        if self.projections_nounits:
            if self.projections:
                # Check if we've already seen/processed these exact projections
                changed_projections = [
                    p
                    for p in self.projections
                    if not any([n for n in self.projections_nounits if n.year == p.year and n.value == p.value.m])
                ]
                if changed_projections:
                    raise ValueError
                return
            self.projections = [
                IProjection(year=p.year, value=BenchmarkQuantity(Q_(p.value, benchmark_metric)))
                for p in self.projections_nounits
                if p.year in range(ProjectionControls.BASE_YEAR, ProjectionControls.TARGET_YEAR + 1)
            ]
        elif not self.projections:
            logger.warning(f"Empty Benchmark for sector {sector}, region {region}")

    def __getitem__(self, item):
        return getattr(self, item)


class IBenchmarks(BaseModel):
    benchmarks: List[IBenchmark]
    production_centric: bool = False

    def __getitem__(self, item):
        return getattr(self, item)


# These IProductionBenchmarkScopes and IEIBenchmarkScopes are vessels for holding initialization data
# The CompanyDataProvider methods create their own dataframes that are then used throughout


class IProductionBenchmarkScopes(BaseModel):
    AnyScope: Optional[IBenchmarks] = None
    S1: Optional[IBenchmarks] = None
    S2: Optional[IBenchmarks] = None
    S1S2: Optional[IBenchmarks] = None
    S3: Optional[IBenchmarks] = None
    S1S2S3: Optional[IBenchmarks] = None

    def __getitem__(self, item):
        return getattr(self, item)


class IEIBenchmarkScopes(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    S1: Optional[IBenchmarks] = None
    S2: Optional[IBenchmarks] = None
    S1S2: Optional[IBenchmarks] = None
    S3: Optional[IBenchmarks] = None
    S1S2S3: Optional[IBenchmarks] = None
    benchmark_temperature: Quantity_type("delta_degC")
    benchmark_global_budget: Quantity_type("Gt CO2")
    is_AFOLU_included: bool

    def __getitem__(self, item):
        return getattr(self, item)


class ICompanyEIProjection(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    year: int
    value: Optional[EI_Quantity] = None

    def __getitem__(self, item):
        return getattr(self, item)

    def __eq__(self, o):
        if self.year != o.year:
            raise ValueError(f"EI Projection years not aligned for __eq__(): {self.year} vs. {o.year}")
        if ITR.isna(self.value.m) and ITR.isna(o.value.m):
            return True
        return self.value == o.value

    def add(self, o):
        if self.year != o.year:
            # breakpoint()
            raise ValueError(f"EI Projection years not aligned for add(): {self.year} vs. {o.year}")
        return ICompanyEIProjection(
            year=self.year,
            value=self.value if ITR.isna(o.value.m) else self.value + o.value.to(self.value.units),
        )

    def min(self, o):
        if self.year != o.year:
            raise ValueError(f"EI Projection years not aligned for min(): {self.year} vs. {o.year}")
        return ICompanyEIProjection(year=self.year, value=min(self.value, o.value))


class ICompanyEIProjections(BaseModel):
    ei_metric: EI_Metric
    projections: List[ICompanyEIProjection]

    def __getitem__(self, item):
        return getattr(self, item)

    def __str__(self):
        # Work-around for https://github.com/hgrecco/pint/issues/1687
        ei_metric = ureg.parse_units(self.ei_metric)
        series = (
            lambda z: (
                idx := z[0],
                values := z[1],
                pd.Series(PA_(np.asarray(values), dtype=str(ei_metric)), index=idx),
            )[-1]
        )(list(zip(*[(x.year, round(ITR.Q_m_as(x.value, ei_metric), 4)) for x in self.projections])))
        return str(series)


class DF_ICompanyEIProjections(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    ei_metric: Optional[EI_Metric] = None
    projections: pd.Series

    @field_validator("projections")
    def val_projections(cls, v: pd.Series):
        if isinstance(v.pint, PintSeriesAccessor):
            return v
        raise ValidationError(f"{v} is not composed of a PintArray")

    def __init__(self, icompany_ei_projections=None, *args, **kwargs):
        projections = None
        projections_gen = None
        if icompany_ei_projections is not None:
            ei_metric = icompany_ei_projections.ei_metric
            projections_gen = icompany_ei_projections.projections
        else:
            ei_metric = kwargs["ei_metric"]
            projections = kwargs["projections"]
            if not isinstance(projections, pd.Series):
                projections_gen = projections
                projections = None
        if projections_gen is not None:
            # Work-around for https://github.com/hgrecco/pint/issues/1687
            ei_metric = ureg.parse_units(ei_metric)
            years, values = list(
                map(
                    list,
                    zip(
                        *[
                            (
                                x["year"],
                                np.nan if x["value"] is None else ITR.Q_m_as(x["value"], ei_metric, inplace=True),
                            )
                            for x in projections_gen
                        ]
                    ),
                )
            )
            projections = pd.Series(
                PA_(np.asarray(values), dtype=str(ei_metric)),
                index=pd.Index(years, name="year"),
                name="value",
            )
        super().__init__(ei_metric=str(ei_metric), projections=projections)


class ICompanyEIProjectionsScopes(BaseModel):
    S1: Optional[DF_ICompanyEIProjections] = None
    S2: Optional[DF_ICompanyEIProjections] = None
    S1S2: Optional[DF_ICompanyEIProjections] = None
    S3: Optional[DF_ICompanyEIProjections] = None
    S1S2S3: Optional[DF_ICompanyEIProjections] = None

    def __init__(self, *args, **kwargs):
        # We don't validate anything in the first step because incoming parameters are the wild west
        # (dict, ICompanyEIProjections, pd.Series)
        super().__init__()

        for k, v in kwargs.items():
            if isinstance(v, dict):
                setattr(
                    self,
                    k,
                    DF_ICompanyEIProjections(
                        ei_metric=EI_Metric(v["ei_metric"]),
                        projections=v["projections"],
                    ),
                )
            elif isinstance(v, ICompanyEIProjections):
                setattr(self, k, DF_ICompanyEIProjections(icompany_ei_projections=v))
            elif isinstance(v, pd.Series):
                ei_metric = EI_Metric(str(v.dtype))
                if ei_metric.startswith("pint["):
                    ei_metric = ei_metric[5:-1]
                setattr(
                    self,
                    k,
                    DF_ICompanyEIProjections(ei_metric=ei_metric, projections=v),
                )
            elif isinstance(v, DF_ICompanyEIProjections) or v is None:
                setattr(self, k, v)
            else:
                # breakpoint()
                assert False
            # We could do a post-hoc validation here...

    def __getitem__(self, item):
        return getattr(self, item)

    def __str__(self):
        dict_items = {
            scope: getattr(self, scope).projections
            for scope in ["S1", "S2", "S1S2", "S3", "S1S2S3"]
            if getattr(self, scope) is not None
        }
        return str(pd.DataFrame.from_dict(dict_items))


class IProductionRealization(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    year: int
    value: Optional[ProductionQuantity] = None


class IEmissionRealization(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    year: int
    value: Optional[EmissionsQuantity] = None

    def __getitem__(self, item):
        return getattr(self, item)

    def __eq__(self, o):
        assert self.year == o.year
        if ITR.isna(self.value.m) and ITR.isna(o.value.m):
            return True
        return self.value == o.value

    def add(self, o):
        assert self.year == o.year
        return IEmissionRealization(
            year=self.year,
            value=self.value if ITR.isna(o.value.m) else self.value + o.value.to(self.value.units),
        )


class IHistoricEmissionsScopes(BaseModel):
    S1: List[IEmissionRealization]
    S2: List[IEmissionRealization]
    S1S2: List[IEmissionRealization]
    S3: List[IEmissionRealization]
    S1S2S3: List[IEmissionRealization]

    def __getitem__(self, item):
        return getattr(self, item)

    def __str__(self):
        dict_items = {
            scope: (
                lambda z: (
                    idx := z[0],
                    values := z[1],
                    pd.Series(PA_(np.asarray(values), dtype="Mt CO2e"), index=idx),
                )[-1]
            )(list(zip(*[(x.year, round(x.value.to("Mt CO2e").m, 4)) for x in getattr(self, scope)])))
            for scope in ["S1", "S2", "S1S2", "S3", "S1S2S3"]
            if getattr(self, scope) is not None
        }
        return str(pd.DataFrame.from_dict(dict_items))


class IEIRealization(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    year: int
    value: Optional[EI_Quantity] = None

    def __getitem__(self, item):
        return getattr(self, item)

    def __eq__(self, o):
        assert self.year == o.year
        if ITR.isna(self.value.m) and ITR.isna(o.value.m):
            return True
        return self.value == o.value

    def add(self, o):
        assert self.year == o.year
        return IEIRealization(
            year=self.year,
            value=self.value if ITR.isna(o.value.m) else self.value + o.value.to(self.value.units),
        )


class IHistoricEIScopes(BaseModel):
    S1: List[IEIRealization]
    S2: List[IEIRealization]
    S1S2: List[IEIRealization]
    S3: List[IEIRealization]
    S1S2S3: List[IEIRealization]

    def __getitem__(self, item):
        return getattr(self, item)

    def __str__(self):
        dict_items = {
            scope: (
                lambda z: (
                    idx := z[0],
                    values := z[1],
                    pd.Series(PA_(np.asarray(values), dtype=ei_metric), index=idx),
                )[-1]
            )(list(zip(*[(x.year, round(x.value.to(ei_metric).m, 4)) for x in getattr(self, scope)])))
            for scope in ["S1", "S2", "S1S2", "S3", "S1S2S3"]
            # Work-around for https://github.com/hgrecco/pint/issues/1687
            for ei_metric in [str(ureg.parse_units(getattr(self, scope).ei_metric))]
            if getattr(self, scope) is not None
        }
        return str(pd.DataFrame.from_dict(dict_items))


class IHistoricData(BaseModel):
    productions: Optional[List[IProductionRealization]] = None
    emissions: Optional[IHistoricEmissionsScopes] = None
    emissions_intensities: Optional[IHistoricEIScopes] = None


class ITargetData(BaseModel):
    netzero_year: Optional[int]
    target_type: Union[
        Literal["intensity"],
        Literal["absolute"],
        Literal["Intensity"],
        Literal["Absolute"],
    ]
    target_scope: EScope
    target_start_year: Optional[int] = None
    target_base_year: int
    target_end_year: int

    target_base_year_qty: float
    target_base_year_err: Optional[float] = None
    target_base_year_unit: str
    target_reduction_pct: float  # This is actually a fraction, not a percentage.  1.0 = complete reduction to zero.

    @model_validator(mode="before")
    def start_end_base_order(cls, v):
        if v["target_start_year"] < v["target_base_year"]:
            raise ValueError(
                f"Scope {v['target_scope']}: Target start year ({v['target_start_year']}) must be equal or greater than base year {v['target_base_year']}"
            )
        if v["target_end_year"] <= v["target_base_year"]:
            raise ValueError(
                f"Scope {v['target_scope']}: Target end year ({v['target_end_year']}) must be greater than base year {v['target_base_year']}"
            )
        if v["target_end_year"] <= v["target_start_year"]:
            raise ValueError(
                f"Scope {v['target_scope']}: Target end year ({v['target_end_year']}) must be greater than start year {v['target_start_year']}"
            )
        return v


class ICompanyData(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    company_name: str
    company_id: str

    sector: str  # TODO: make SortableEnums
    region: str  # TODO: make SortableEnums
    # TemperatureScoreConfig.CONTROLS_CONFIG.target_probability is not company-specific,
    # while target_probability in ICompanyData is company-specific
    target_probability: float = np.nan

    target_data: Optional[List[ITargetData]] = None
    # IHistoric data can contain None values; need to convert to Quantified NaNs
    historic_data: Optional[IHistoricData] = None

    country: Optional[str] = None

    # Emissions typically use t CO2 for MWh/GJ and Mt CO2 for TWh/PJ
    emissions_metric: Optional[EmissionsMetric] = None
    production_metric: Optional[ProductionMetric] = None

    # These three instance variables match against financial data below, but are incomplete as historic_data and target_data
    base_year_production: Optional[ProductionQuantity] = None
    ghg_s1s2: Optional[EmissionsQuantity] = None
    ghg_s3: Optional[EmissionsQuantity] = None

    industry_level_1: Optional[str] = None
    industry_level_2: Optional[str] = None
    industry_level_3: Optional[str] = None
    industry_level_4: Optional[str] = None

    company_revenue: Optional[MonetaryQuantity] = None
    company_market_cap: Optional[MonetaryQuantity] = None
    company_enterprise_value: Optional[MonetaryQuantity] = None
    company_ev_plus_cash: Optional[MonetaryQuantity] = None
    company_total_assets: Optional[MonetaryQuantity] = None
    company_cash_equivalents: Optional[MonetaryQuantity] = None

    # Initialized later when we have benchmark information.  It is OK to initialize as None and fix later.
    # They will show up as {'S1S2': { 'projections': [ ... ] }}
    projected_targets: Optional[ICompanyEIProjectionsScopes] = None
    projected_intensities: Optional[ICompanyEIProjectionsScopes] = None

    # TODO: Do we want to do some sector inferencing here?

    def _sector_to_production_units(self, sector, region="Global"):
        sector_unit_dict = {
            "Electricity Utilities": {"North America": "MWh", "Global": "GJ"},
            "Gas Utilities": {"Global": "PJ"},
            "Utilities": {"Global": "PJ"},
            "Steel": {"Global": "t Steel"},
            "Aluminum": {"Global": "t Aluminum"},
            "Energy": {"Global": "PJ"},
            "Coal": {"Global": "t Coal"},
            "Oil": {"Global": "bbl/d"},
            "Gas": {"Global": "bcm"},
            "Oil & Gas": {"Global": "PJ"},
            "Autos": {"Global": "pkm"},
            "Trucking": {"Global": "tkm"},
            "Cement": {"Global": "t Cement"},
            "Construction Buildings": {"Global": "billion USD"},
            "Residential Buildings": {"Global": "billion m**2"},  # Should it be 'built m**2' ?
            "Commercial Buildings": {"Global": "billion m**2"},  # Should it be 'built m**2' ?
            "Textiles": {"Global": "billion USD"},
            "Chemicals": {"Global": "billion USD"},
            "Pharmaceuticals": {"Global": "billion USD"},
            "Ag Chem": {"Global": "billion USD"},
            "Consumer Products": {"Global": "billion USD"},
            "Fiber & Rubber": {"Global": "billion USD"},
            "Petrochem & Plastics": {"Global": "billion USD"},
        }
        units = None
        if sector_unit_dict.get(sector):
            region_unit_dict = sector_unit_dict[sector]
            if region_unit_dict.get(region):
                units = region_unit_dict[region]
            else:
                units = region_unit_dict["Global"]
        else:
            raise ValueError(f"No source of production metrics for {self.company_name}")
        return units

    def _get_base_realization_from_historic(self, realized_values: List[BaseModel], metric, base_year=None):
        valid_realizations = [rv for rv in realized_values if not ITR.isna(rv.value)]
        if not valid_realizations:
            retval = realized_values[0].model_copy()
            retval.year = None
            return retval
        valid_realizations.sort(key=lambda x: x.year, reverse=True)
        if base_year and valid_realizations[0].year != base_year:
            retval = realized_values[0].copy()
            retval.year = base_year
            # FIXME: Unless and until we accept uncertainties as input, rather than computed data, we don't need to make this a UFloat here
            retval.value = PintType(metric).na_value
            return retval
        return valid_realizations[0]

    def _normalize_historic_data(
        self,
        historic_data: IHistoricData,
        production_metric: ProductionMetric,
        emissions_metric: EmissionsMetric,
    ) -> IHistoricData:
        def _normalize(value, metric):
            if value is not None:
                if value.u == metric:
                    return value
                if ITR.isna(value):
                    return PintType(metric).na_value
                # We've pre-conditioned metric so don't need to work around https://github.com/hgrecco/pint/issues/1687
                return value.to(metric)
            return PintType(metric).na_value

        if historic_data is None:
            return None

        if historic_data.productions:
            # Work-around for https://github.com/hgrecco/pint/issues/1687
            production_metric = ureg.parse_units(production_metric)  # Catch things like '$'
            historic_data.productions = [
                IProductionRealization(year=p.year, value=_normalize(p.value, production_metric))
                for p in historic_data.productions
            ]
        # Work-around for https://github.com/hgrecco/pint/issues/1687
        # emissions_metric = ureg(emissions_metric).u
        ei_metric = ureg.parse_units(f"{emissions_metric} / ({production_metric})")
        for scope_name in EScope.get_scopes():
            if historic_data.emissions:
                setattr(
                    historic_data.emissions,
                    scope_name,
                    [
                        IEmissionRealization(year=p.year, value=_normalize(p.value, emissions_metric))
                        for p in historic_data.emissions[scope_name]
                    ],
                )
            if historic_data.emissions_intensities:
                setattr(
                    historic_data.emissions_intensities,
                    scope_name,
                    [
                        IEIRealization(year=p.year, value=_normalize(p.value, ei_metric))
                        for p in historic_data.emissions_intensities[scope_name]
                    ],
                )
        return historic_data

    def __init__(
        self,
        emissions_metric=None,
        production_metric=None,
        base_year_production=None,
        ghg_s1s2=None,
        ghg_s3=None,
        target_data=None,
        historic_data=None,
        *args,
        **kwargs,
    ):
        super().__init__(
            emissions_metric=emissions_metric,
            production_metric=production_metric,
            base_year_production=base_year_production,
            ghg_s1s2=ghg_s1s2,
            ghg_s3=ghg_s3,
            target_data=target_data,
            historic_data=historic_data,
            *args,
            **kwargs,
        )
        # In-bound parameters are JSON (str, int, float, dict), which are converted to models by __super__ and stored as instance variables
        if production_metric is None:
            units = self._sector_to_production_units(self.sector, self.region)
            self.production_metric = ProductionMetric(units)
            if emissions_metric is None:
                self.emissions_metric = EmissionsMetric("t CO2")
        elif emissions_metric is None:
            if str(self.production_metric) in [
                "TWh",
                "PJ",
                "Mt Steel",
                "megaFe_ton",
                "mmboe",
            ]:
                self.emissions_metric = EmissionsMetric("Mt CO2")
            else:
                self.emissions_metric = EmissionsMetric("t CO2")
            # TODO: Should raise a warning here

        # This is only a partial initialization
        if self.historic_data is None:
            return
        self.historic_data = self._normalize_historic_data(
            self.historic_data, self.production_metric, self.emissions_metric
        )
        base_year = None
        if self.base_year_production:
            pass
        # Right now historic_data comes in via template.py ESG data
        elif self.historic_data.productions:
            # TODO: This is a hack to get things going.
            base_realization = self._get_base_realization_from_historic(
                self.historic_data.productions, self.production_metric, base_year
            )
            base_year = base_realization.year
            self.base_year_production = base_realization.value
        else:
            logger.warning(f"missing historic data for base_year_production for {self.company_name}")
            self.base_year_production = PintType(self.production_metric).na_value
        if self.ghg_s1s2 is None and self.historic_data.emissions:
            if self.historic_data.emissions.S1S2:
                base_realization = self._get_base_realization_from_historic(
                    self.historic_data.emissions.S1S2, self.emissions_metric, base_year
                )
                base_year = base_year or base_realization.year
                self.ghg_s1s2 = base_realization.value
            elif self.historic_data.emissions.S1 and self.historic_data.emissions.S2:
                base_realization_s1 = self._get_base_realization_from_historic(
                    self.historic_data.emissions.S1, self.emissions_metric, base_year
                )
                base_realization_s2 = self._get_base_realization_from_historic(
                    self.historic_data.emissions.S2, self.emissions_metric, base_year
                )
                base_year = base_year or base_realization_s1.year
                if base_realization_s1.value is not None and base_realization_s2.value is not None:
                    self.ghg_s1s2 = base_realization_s1.value + base_realization_s2.value
        if self.ghg_s1s2 is None and self.historic_data.emissions_intensities:
            intensity_metric = ureg.parse_units(f"({self.emissions_metric}) / ({self.production_metric})")
            if self.historic_data.emissions_intensities.S1S2:
                base_realization = self._get_base_realization_from_historic(
                    self.historic_data.emissions_intensities.S1S2,
                    intensity_metric,
                    base_year,
                )
                base_year = base_year or base_realization.year
                if base_realization.value is not None:
                    self.ghg_s1s2 = base_realization.value * self.base_year_production
            elif self.historic_data.emissions_intensities.S1 and self.historic_data.emissions_intensities.S2:
                base_realization_s1 = self._get_base_realization_from_historic(
                    self.historic_data.emissions_intensities.S1,
                    intensity_metric,
                    base_year,
                )
                base_realization_s2 = self._get_base_realization_from_historic(
                    self.historic_data.emissions_intensities.S2,
                    intensity_metric,
                    base_year,
                )
                base_year = base_year or base_realization_s1.year
                if base_realization_s1.value is not None and base_realization_s2.value is not None:
                    self.ghg_s1s2 = (base_realization_s1.value + base_realization_s2.value) * self.base_year_production
            else:
                raise ValueError(f"missing S1S2 historic intensity data for {self.company_name}")
        if self.ghg_s1s2 is None:
            raise ValueError(
                f"missing historic emissions or intensity data to calculate ghg_s1s2 for {self.company_name}"
            )
        if self.ghg_s3 is None and self.historic_data.emissions and self.historic_data.emissions.S3:
            base_realization_s3 = self._get_base_realization_from_historic(
                self.historic_data.emissions.S3, self.emissions_metric, base_year
            )
            self.ghg_s3 = base_realization_s3.value
        if self.ghg_s3 is None and self.historic_data.emissions_intensities:
            if self.historic_data.emissions_intensities.S3:
                intensity_metric = ureg.parse_units(f"({self.emissions_metric}) / ({self.production_metric})")
                base_realization_s3 = self._get_base_realization_from_historic(
                    self.historic_data.emissions_intensities.S3,
                    intensity_metric,
                    base_year,
                )
                if base_realization_s3.value is not None:
                    self.ghg_s3 = base_realization_s3.value * self.base_year_production


# These aggregate terms are all derived from the benchmark being used
class ICompanyAggregates(ICompanyData):
    cumulative_budget: Optional[EmissionsQuantity] = None
    cumulative_scaled_budget: Optional[EmissionsQuantity] = None
    cumulative_trajectory: Optional[EmissionsQuantity] = None
    cumulative_target: Optional[EmissionsQuantity] = None
    benchmark_temperature: Optional[Quantity_type("delta_degC")] = None
    benchmark_global_budget: Optional[EmissionsQuantity] = None
    scope: Optional[EScope] = None

    # The first year that cumulative_projections exceeds the 2050 cumulative_budget
    trajectory_exceedance_year: Optional[int] = None
    target_exceedance_year: Optional[int] = None

    # projected_production is computed but never saved, so computed at least 2x: initialiation/projection and cumulative budget
    # projected_targets: Optional[ICompanyEIProjectionsScopes] = None
    # projected_intensities: Optional[ICompanyEIProjectionsScopes] = None

    # Custom validator here
    @field_validator("trajectory_exceedance_year", "target_exceedance_year")
    def val_exceedance_year(cls, v):
        if isinstance(v, int):
            return v
        if pd.isna(v):
            return None
        raise ValueError(f"{v} is not compatible with Int64 dtype")

    @classmethod
    def from_ICompanyData(cls, super_instance, scope_company_data):
        """
        Fast way to add instance variables to a pre-validated SUPER_INSTANCE
        SCOPE_COMPANY_DATA is the dictionary of the new values we want to add...for this one company
        """
        # FIXME: Would love to know how to run these automatically...
        EmissionsQuantity(scope_company_data["cumulative_budget"])
        EmissionsQuantity(scope_company_data["cumulative_scaled_budget"])
        if not ITR.isna(scope_company_data["cumulative_trajectory"]):
            EmissionsQuantity(scope_company_data["cumulative_trajectory"])
        if not ITR.isna(scope_company_data["cumulative_target"]):
            EmissionsQuantity(scope_company_data["cumulative_target"])
        if not Q_(scope_company_data["benchmark_temperature"]).is_compatible_with(ureg("delta_degC")):
            raise ValueError(
                f"benchmark temperature {scope_company_data['benchmark_temperature']} is not compatible with delta_degC"
            )
        else:
            scope_company_data["benchmark_temperature"] = Q_(scope_company_data["benchmark_temperature"])
        EmissionsQuantity(scope_company_data["benchmark_global_budget"])
        if not isinstance(scope_company_data["scope"], EScope):
            raise ValueError(f"scope {scope_company_data['scope']} is not a valid scope")
        if not ITR.isna(scope_company_data["trajectory_exceedance_year"]) and not isinstance(
            scope_company_data["trajectory_exceedance_year"], int
        ):
            raise ValueError(
                f"scope {scope_company_data['trajectory_exceedance_year']} is not a valid trajectory exceedance year value"
            )
        if not ITR.isna(scope_company_data["target_exceedance_year"]) and not isinstance(
            scope_company_data["target_exceedance_year"], int
        ):
            raise ValueError(
                f"scope {scope_company_data['target_exceedance_year']} is not a valid target exceedance year value"
            )
        # ...while not re-running any validation on super_instnace
        return cls.model_construct(**scope_company_data, **super_instance.__dict__)



import logging
import os
import pathlib
from typing import List, Type

import numpy as np
import osc_ingest_trino as osc
import pandas as pd
import sqlalchemy
from dotenv import load_dotenv
from pint_pandas import PintArray

from ..configs import ColumnsConfig, LoggingConfig
from ..data import ureg

# Rather than duplicating a few methods from BaseCompanyDataProvider, we just call them to delegate to them
from ..data.base_providers import BaseCompanyDataProvider
from ..data.data_providers import (
    CompanyDataProvider,
    IntensityBenchmarkDataProvider,
    ProductionBenchmarkDataProvider,
)
from ..data.data_warehouse import DataWarehouse
from ..data.osc_units import Quantity
from ..interfaces import (
    EScope,
    IBenchmark,
    ICompanyAggregates,
    ICompanyData,
    IEIBenchmarkScopes,
    IProductionBenchmarkScopes,
)

# TODO handle ways to append information (from other providers, other benchmarks, new scope info, new corp data updates, etc)


logger = logging.getLogger(__name__)
LoggingConfig.add_config_to_logger(logger)

# Load some standard environment variables from a dot-env file, if it exists.
# If no such file can be found, does not fail, and so allows these environment vars to
# be populated in some other way
dotenv_dir = os.environ.get("CREDENTIAL_DOTENV_DIR", os.environ.get("HOME", "/opt/app-root/src"))
dotenv_path = pathlib.Path(dotenv_dir) / "credentials.env"
if os.path.exists(dotenv_path):
    load_dotenv(dotenv_path=dotenv_path, override=True)

ingest_catalog = "osc_datacommons_dev"
ingest_schema = "demo_dv"
demo_schema = "demo_dv"

engine = osc.attach_trino_engine(verbose=True, catalog=ingest_catalog, schema=ingest_schema)


# If DF_COL contains Pint quantities (because it is a PintArray or an array of Pint Quantities),
# return a two-column dataframe of magnitudes and units.
# If DF_COL contains no Pint quanities, return it unchanged.
def dequantify_column(df_col: pd.Series) -> pd.DataFrame:
    if isinstance(df_col.values, PintArray):
        return pd.DataFrame(
            {
                df_col.name: df_col.values.quantity.m,
                df_col.name + "_units": str(df_col.values.dtype.units),
            },
            index=df_col.index,
        )
    elif df_col.size == 0:
        return df_col
    elif isinstance(df_col.iloc[0], Quantity):
        m, u = list(zip(*df_col.map(lambda x: (np.nan, "dimensionless") if pd.isna(x) else (x.m, str(x.u)))))
        return pd.DataFrame({df_col.name: m, df_col.name + "_units": u}, index=df_col.index)
    else:
        return df_col


# Rewrite dataframe DF so that columns containing Pint quantities are represented by a column for the Magnitude and column for the Units.
# The magnitude column retains the original column name and the units column is renamed with a _units suffix.
def dequantify_df(df: pd.DataFrame) -> pd.DataFrame:
    return pd.concat([dequantify_column(df[col]) for col in df.columns], axis=1)


# Because this DF comes from reading a Trino table, and because columns must be unqiue, we don't have to enumerate to ensure we properly handle columns with duplicated names
def requantify_df(df: pd.DataFrame, typemap={}) -> pd.DataFrame:
    units_col = None
    columns_not_found = [k for k in typemap.keys() if k not in df.columns]
    if columns_not_found:
        logger.error(f"columns {columns_not_found} not found in DataFrame")
        raise ValueError
    columns_reversed = reversed(df.columns)
    for col in columns_reversed:
        if col.endswith("_units"):
            if units_col:
                logger.error(f"Column {units_col} follows {col} without intervening value column")
                # We expect _units column to follow a non-units column
                raise ValueError
            units_col = col
            continue
        if units_col:
            if col + "_units" != units_col:
                logger.error(f"Excpecting column name {col}_units but saw {units_col} instead")
                raise ValueError
            if (df[units_col] == df[units_col].iloc[0]).all():
                # We can make a PintArray since column is of homogeneous type
                # ...and if the first valid index matches all, we can take first row as good
                new_col = PintArray(df[col], dtype=f"pint[{ureg(df[units_col].iloc[0]).u}]")
            else:
                # Make a pd.Series of Quantity in a way that does not throw UnitStrippedWarning
                new_col = pd.Series(data=df[col], name=col) * pd.Series(
                    data=df[units_col].map(lambda x: typemap.get(col, "dimensionless") if pd.isna(x) else ureg(x).u),
                    name=col,
                )
            if col in typemap.keys():
                new_col = new_col.astype(f"pint[{typemap[col]}]")
            df = df.drop(columns=units_col)
            df[col] = new_col
            units_col = None
        elif col in typemap.keys():
            df[col] = df[col].astype(f"pint[{typemap[col]}]")
    return df


def create_table_from_df(
    df: pd.DataFrame,
    schemaname: str,
    tablename: str,
    engine: sqlalchemy.engine.base.Engine,
    verbose=False,
):
    drop_table = f"drop table if exists {schemaname}.{tablename}"
    qres = osc._do_sql(drop_table, engine, verbose)
    logger.debug("dtypes, columns, and index of create_table_from_df(df...)")
    logger.debug(df.dtypes)
    logger.debug(df.columns)
    logger.debug(df.index)
    new_df = dequantify_df(df)
    new_df.to_sql(
        tablename,
        con=engine,
        schema=schemaname,
        if_exists="append",
        index=False,
        method=osc.TrinoBatchInsert(batch_size=5000, verbose=True),
    )


# When reading SQL tables to import into DataFrames, it is up to the user to preserve {COL}, {COL}_units pairings so they can be reconstructed.
# If the user does a naive "select * from ..." this happens naturally.
# We can give a warning when we see a resulting dataframe that could have, but does not have, unit information properly integrated.  But
# fixing the query on the fly becomes difficult when we consider the fully complexity of parsing and rewriting SQL queries to put the units columns in the correct locations.
# (i.e., properly in the principal SELECT clause (which can have arbitrarily complex terms), not confused by FROM, WHERE, GROUP BY, ORDER BY, etc.)


def read_quantified_sql(
    sql: str,
    tablename,
    schemaname,
    engine: sqlalchemy.engine.base.Engine,
    index_col=None,
) -> pd.DataFrame:
    qres = osc._do_sql(f"describe {schemaname}.{tablename}", engine, verbose=False)
    # tabledesc will be a list of tuples (column, type, extra, comment)
    colnames = [x[0] for x in qres]
    # read columns normally...this will be missing any unit-related information
    sql_df = pd.read_sql(sql, engine, index_col)
    # if the query requests columns that don't otherwise bring unit information along with them, get that information too
    extra_unit_columns = [
        (i, f"{col}_units")
        for i, col in enumerate(sql_df.columns)
        if f"{col}_units" not in sql_df.columns and f"{col}_units" in colnames
    ]
    if extra_unit_columns:
        extra_unit_columns_positions = [
            (i, extra_unit_columns[i][0], extra_unit_columns[i][1]) for i in range(len(extra_unit_columns))
        ]
        for col_tuple in extra_unit_columns_positions:
            logger.error(
                f"Missing units column '{col_tuple[2]}' after original column '{sql_df.columns[col_tuple[1]]}' (should be column #{col_tuple[0]+col_tuple[1]+1} in new query)"
            )
        raise ValueError
    else:
        return requantify_df(sql_df).convert_dtypes()


# Basic Corp Data Asumptions
#   5 year historical EI (else we presume single year is constant backward and forward)
#   5 year historical Production (else we presume single year is constant backward and forward)
#   5 year historical Emissions (else we presume single year is constant backward and forward)
#   We can infer one of the above from the other two (simple maths)
#   The above tables identify the scope(s) to which they apply (S1, S2, S12, S3, S123) and data source (e.g. 'rmi_20211120')

# Basic Benchmark Data Assumptions
#   EI for a given scope
#   Production defined in terms of growth (or negative growth) on a rolling basis (so 0.05, -0.04) would mean 5% growth followed by 4% negative growth for a total of 0.8%
#   Benchmarks are named (e.g., 'OECM')


class VaultCompanyDataProvider(CompanyDataProvider):
    """
    This class serves primarily for connecting to the ITR tool to the Data Vault via Trino.

    :param company_table: the name of the Trino table that contains fundamental data for companies
    :param target_table: the name of the Trino table that contains company (emissions intensity) target data (and possibly historical data)
    :param trajectory_table: the name of the Trino table that contains company (emissions intensity) historical data (and possibly trajectory data)
    :param company_schema: the name of the schema where the company_table is found
    :param column_config: An optional ColumnsConfig object containing relevant variable names
    """

    def __init__(
        self,
        engine: sqlalchemy.engine.base.Engine,
        company_table: str,
        target_table: str = None,
        trajectory_table: str = None,
        company_schema: str = None,
        column_config: Type[ColumnsConfig] = ColumnsConfig,
    ):
        super().__init__()
        self._engine = engine
        self._schema = company_schema or engine.dialect.default_schema_name or "demo_dv"
        self._company_table = company_table
        self.column_config = column_config
        # Validate and complete the projected trajectories
        self._target_table = target_table or company_table.replace("company_", "target_")  # target_data
        self._trajectory_table = trajectory_table or company_table.replace("company_", "trajectory_")  # trajectory_data
        self._production_table = company_table.replace("company_", "production_")  # production_data
        self._emissions_table = company_table.replace("company_", "emissions_")  # emissions_data
        companies_without_projections = osc._do_sql(
            f"""
select C.company_name, C.company_id from {self._schema}.{self._company_table} C left join {self._schema}.{self._target_table} EI on EI.company_id=C.company_id
where EI.ei_s1_by_year is NULL and EI.ei_s1s2_by_year is NULL and EI.ei_s1s2s3_by_year is NULL
""",
            self._engine,
            verbose=True,
        )
        if companies_without_projections:
            logger.error(
                f"Provide either historic emissions data or projections for companies with IDs {companies_without_projections}"
            )

    # The factors one would want to sum over companies for weighting purposes are:
    #   * market_cap_usd
    #   * enterprise_value_usd
    #   * assets_usd
    #   * revenue_usd
    #   * emissions

    # TODO: make return value a Quantity (USD or CO2)
    def sum_over_companies(
        self,
        company_ids: List[str],
        year: int,
        factor: str,
        scope: EScope = EScope.S1S2,
    ) -> float:
        if factor == "enterprise_value_usd":
            qres = osc._do_sql(
                f"select sum(market_cap_usd + debt_usd - cash_usd) as {factor}_sum from {self._schema}.{self._company_table} where year={year}",
                self._engine,
                verbose=False,
            )
        elif factor == "emissions":
            if scope in [EScope.S1, EScope.S2, EScope.S3]:
                qres = osc._do_sql(
                    f"select sum(co2_{scope.name.lower()}_by_year) as {factor}_sum from {self._schema}.{self._emissions_table} where year={year}",
                    self._engine,
                    verbose=False,
                )
            elif scope == EScope.S1S2:
                qres = osc._do_sql(
                    f"select sum(co2_s1_by_year+if(is_nan(co2_s2_by_year),0.0,co2_s2_by_year)) as {factor}_sum from {self._schema}.{self._emissions_table} where year={year}",
                    self._engine,
                    verbose=False,
                )
            elif scope == EScope.S1S2S3:
                qres = osc._do_sql(
                    f"select sum(co2_s1_by_year+if(is_nan(co2_s2_by_year),0.0,co2_s2_by_year)+if(is_nan(co2_s3_by_year),0.0,co2_s3_by_year)) as {factor}_sum from {self._schema}.{self._emissions_table} where year={year}",
                    self._engine,
                    verbose=False,
                )
            else:
                assert False
        else:
            qres = osc._do_sql(
                f"select sum({factor}) as {factor}_sum from {self._schema}.{self._company_table} where year={year}",
                self._engine,
                verbose=False,
            )
        # qres[0] is the first row of the returned data; qres[0][0] is the first (and only) column of the row returned
        return qres[0][0]

    def compute_portfolio_weights(
        self,
        pa_temp_scores: pd.Series,
        year: int,
        factor: str,
        scope: EScope = EScope.S1S2,
    ) -> pd.Series:
        """
        Portfolio values could be position size, temperature scores, anything that can be multiplied by a factor.

        :param company_ids: A pd.Series of company IDs (ISINs)
        :return: A pd.Series weighted by the factor
        """
        if factor == "company_evic":
            qres = osc._do_sql(
                f"select company_id, sum(company_market_cap + company_cash_equivalents) as {factor} from {self._schema}.{self._company_table} group by company_id",
                self._engine,
                verbose=False,
            )
        elif factor == "emissions":
            if scope in [EScope.S1, EScope.S2, EScope.S3]:
                qres = osc._do_sql(
                    f"select company_id, sum(co2_{scope.name.lower()}_by_year) as {factor} from {self._schema}.{self._emissions_table} where year={year} group by company_id",
                    self._engine,
                    verbose=False,
                )
            elif scope == EScope.S1S2:
                qres = osc._do_sql(
                    f"select company_id, sum(co2_s1_by_year+if(is_nan(co2_s2_by_year),0.0,co2_s2_by_year)) as {factor} from {self._schema}.{self._emissions_table} where year={year} group by company_id",
                    self._engine,
                    verbose=False,
                )
            elif scope == EScope.S1S2:
                qres = osc._do_sql(
                    f"select company_id, sum(co2_s1_by_year+if(is_nan(co2_s2_by_year),0.0,co2_s2_by_year)+if(is_nan(co2_s3_by_year),0.0,co2_s3_by_year)) as {factor} from {self._schema}.{self._emissions_table} where year={year} group by company_id",
                    self._engne,
                    verbose=False,
                )
            else:
                assert False
        else:
            qres = osc._do_sql(
                f"select company_id, sum({factor}) as {factor} from {self._schema}.{self._company_table} group by company_id",
                self._engine,
                verbose=False,
            )
        weights = pd.Series(data=[s[1] for s in qres], index=[s[0] for s in qres], dtype=float)
        weights = weights.loc[pa_temp_scores.index.intersection(weights.index)]
        weight_sum = weights.sum()
        return pa_temp_scores * weights / weight_sum

    def get_company_data(self, company_ids: List[str]) -> List[ICompanyData]:
        """
        Get all relevant data for a list of company ids. This method should return a list of ICompanyData
        instances.

        :param company_ids: A list of company IDs (ISINs)
        :return: A list containing the company data
        """
        raise NotImplementedError

    def get_value(self, company_ids: List[str], variable_name: str) -> pd.Series:
        """
        Gets the value of a variable for a list of companies ids
        :param company_ids: list of company ids
        :param variable_name: variable name of the projected feature
        :return: series of values
        """
        raise NotImplementedError

    def get_company_intensity_and_production_at_base_year(self, company_ids: List[str]) -> pd.DataFrame:
        """
        overrides subclass method
        :param: company_ids: list of company ids
        :return: DataFrame the following columns :
        ColumnsConfig.COMPANY_ID, ColumnsConfig.GHG_S1S2, ColumnsConfig.BASE_EI, ColumnsConfig.SECTOR and
        ColumnsConfig.REGION
        """
        raise NotImplementedError

    def get_company_fundamentals(self, company_ids: List[str]) -> pd.DataFrame:
        """
        :param company_ids: A list of company IDs
        :return: A pandas DataFrame with company fundamental info per company
        """
        or_clause = " or ".join([f"company_id = '{c}'" for c in company_ids])
        sql = f"select * from {self._schema}.{self._company_table} where {or_clause}"
        df = read_quantified_sql(sql, self._company_table, self._schema, self._engine)
        # df = df.drop(columns=['projected_targets', 'projected_intensities'])
        return df

    def get_company_projected_trajectories(self, company_ids: List[str]) -> pd.DataFrame:
        """
        :param company_ids: A list of company IDs
        :return: A pandas DataFrame with projected intensities per company
        """
        raise NotImplementedError

    def get_company_projected_targets(self, company_ids: List[str]) -> pd.DataFrame:
        """
        :param company_ids: A list of company IDs
        :return: A pandas DataFrame with projected targets per company
        """
        raise NotImplementedError


class VaultProviderProductionBenchmark(ProductionBenchmarkDataProvider):
    def __init__(
        self,
        engine: sqlalchemy.engine.base.Engine,
        benchmark_name: str,
        production_benchmarks: IProductionBenchmarkScopes,
        ingest_schema: str = None,
        column_config: Type[ColumnsConfig] = ColumnsConfig,
    ):
        """
        Base provider that relies on pydantic interfaces. Default for FastAPI usage
        :param benchmark_name: the table name of the benchmark (in Trino)
        :param production_benchmarks: List of IBenchmarkScopes
        :param column_config: An optional ColumnsConfig object containing relevant variable names
        """
        super().__init__(production_benchmarks=production_benchmarks, column_config=column_config)
        self._engine = engine
        self._schema = ingest_schema or engine.dialect.default_schema_name or "demo_dv"
        self.benchmark_name = benchmark_name
        qres = osc._do_sql(
            f"drop table if exists {self._schema}.{benchmark_name}",
            self._engine,
            verbose=False,
        )
        df = pd.DataFrame()
        for scope in ["AnyScope"]:
            if production_benchmarks.dict()[scope] is None:
                continue
            for benchmark in production_benchmarks.dict()[scope]["benchmarks"]:
                bdf = pd.DataFrame.from_dict(
                    {
                        r["year"]: [
                            r["value"],
                            benchmark["region"],
                            benchmark["sector"],
                            scope,
                        ]
                        for r in benchmark["projections"]
                    },
                    columns=["production", "region", "sector", "scope"],
                    orient="index",
                )
                df = pd.concat([df, bdf])
        df.reset_index(inplace=True)
        df.rename(columns={"index": "year"}, inplace=True)
        df = df.convert_dtypes()
        create_table_from_df(df, self._schema, benchmark_name, engine)

    def get_company_projected_production(self, *args, **kwargs):
        return BaseCompanyDataProvider.get_company_projected_production(*args, **kwargs)

    def get_benchmark_projections(self, *args, **kwargs):
        return BaseCompanyDataProvider.get_benchmark_projections(*args, **kwargs)


class VaultProviderIntensityBenchmark(IntensityBenchmarkDataProvider):
    def __init__(
        self,
        engine: sqlalchemy.engine.base.Engine,
        benchmark_name: str,
        EI_benchmarks: IEIBenchmarkScopes,
        ingest_schema: str = None,
        column_config: Type[ColumnsConfig] = ColumnsConfig,
    ):
        super().__init__(
            EI_benchmarks.benchmark_temperature,
            EI_benchmarks.benchmark_global_budget,
            EI_benchmarks.is_AFOLU_included,
        )
        self._engine = engine
        self._schema = ingest_schema or engine.dialect.default_schema_name or "demo_dv"
        self.benchmark_name = benchmark_name
        df = pd.DataFrame()
        for scope in EScope.get_scopes():
            if EI_benchmarks.dict()[scope] is None:
                continue
            for benchmark in EI_benchmarks.dict()[scope]["benchmarks"]:
                bdf = pd.DataFrame.from_dict(
                    {
                        r["year"]: [
                            r["value"],
                            benchmark["region"],
                            benchmark["sector"],
                            scope,
                            EI_benchmarks.benchmark_global_budget,
                            EI_benchmarks.benchmark_temperature,
                        ]
                        for r in benchmark["projections"]
                    },
                    columns=[
                        "intensity",
                        "region",
                        "sector",
                        "scope",
                        "global_budget",
                        "benchmark_temp",
                    ],
                    orient="index",
                )
                # TODO: AFOLU correction
                df = pd.concat([df, bdf])
        df.reset_index(inplace=True)
        df.rename(columns={"index": "year"}, inplace=True)
        df = df.convert_dtypes()
        create_table_from_df(df, self._schema, benchmark_name, engine)

    def get_SDA_intensity_benchmarks(self, company_info_at_base_year: pd.DataFrame) -> pd.DataFrame:
        """
        Overrides subclass method
        returns a Dataframe with intensity benchmarks per company_id given a region and sector.
        :param benchmark_name: the table name of the benchmark (in Trino)
        :param company_info_at_base_year: DataFrame with at least the following columns :
        ColumnsConfig.COMPANY_ID, ColumnsConfig.BASE_EI ColumnsConfig.SECTOR and ColumnsConfig.REGION
        :return: A DataFrame with company and SDA intensity benchmarks per calendar year per row
        """
        intensity_benchmarks = self._get_intensity_benchmarks(company_info_at_base_year)
        decarbonization_paths = self._get_decarbonizations_paths(intensity_benchmarks)
        last_ei = intensity_benchmarks[self.temp_config.CONTROLS_CONFIG.target_end_year]
        ei_base = company_info_at_base_year[self.column_config.BASE_EI]

        return decarbonization_paths.mul((ei_base - last_ei), axis=0).add(last_ei, axis=0)

    def _get_decarbonizations_paths(self, intensity_benchmarks: pd.DataFrame) -> pd.DataFrame:
        """
        Overrides subclass method
        Returns a DataFrame with the projected decarbonization paths for the supplied companies in intensity_benchmarks.
        :param: A DataFrame with company and intensity benchmarks per calendar year per row
        :return: A pd.DataFrame with company and decarbonisation path s per calendar year per row
        """
        return intensity_benchmarks.apply(lambda row: self._get_decarbonization(row), axis=1)

    def _get_decarbonization(self, intensity_benchmark_row: pd.Series) -> pd.Series:
        """
        Overrides subclass method
        returns a Series with the decarbonization path for a benchmark.
        :param: A Series with company and intensity benchmarks per calendar year per row
        :return: A pd.Series with company and decarbonisation path s per calendar year per row
        """
        first_ei = intensity_benchmark_row[self.temp_config.CONTROLS_CONFIG.base_year]
        last_ei = intensity_benchmark_row[self.temp_config.CONTROLS_CONFIG.target_end_year]
        return intensity_benchmark_row.apply(lambda x: (x - last_ei) / (first_ei - last_ei))

    def _convert_benchmark_to_series(self, benchmark: IBenchmark) -> pd.Series:
        """
        extracts the company projected intensities or targets for a given scope
        :param feature: PROJECTED_EI or PROJECTED_TARGETS
        :param scope: a scope
        :return: pd.Series
        """
        return pd.Series(
            {r.year: r.value for r in benchmark.projections},
            name=(benchmark.region, benchmark.sector),
        )

    def _get_projected_intensities(self, scope: EScope = EScope.S1S2) -> pd.Series:
        """
        Converts IBenchmarkScopes into dataframe for a scope
        :param scope: a scope
        :return: pd.Series
        """
        result = []
        for bm in self._EI_benchmarks.dict()[str(scope)]["benchmarks"]:
            result.append(self._convert_benchmark_to_series(IBenchmark.model_validate(bm)))
        df_bm = pd.DataFrame(result)
        df_bm.index.names = [self.column_config.REGION, self.column_config.SECTOR]

        return df_bm

    def _get_intensity_benchmarks(
        self, company_sector_region_info: pd.DataFrame, scope: EScope = EScope.S1S2
    ) -> pd.DataFrame:
        """
        Overrides subclass method
        returns a Dataframe with production benchmarks per company_id given a region and sector.
        :param company_sector_region_info: DataFrame with at least the following columns :
        ColumnsConfig.COMPANY_ID, ColumnsConfig.SECTOR and ColumnsConfig.REGION
        :param scope: a scope
        :return: A DataFrame with company and intensity benchmarks per calendar year per row
        """
        benchmark_projection = self._get_projected_intensities(scope)  # TODO optimize performance
        sectors = company_sector_region_info[self.column_config.SECTOR]
        regions = company_sector_region_info[self.column_config.REGION]
        benchmark_regions = regions.copy()
        mask = benchmark_regions.isin(benchmark_projection.reset_index()[self.column_config.REGION])
        benchmark_regions.loc[~mask] = "Global"

        benchmark_projection = benchmark_projection.loc[
            list(zip(benchmark_regions, sectors)),
            range(
                self.temp_config.CONTROLS_CONFIG.base_year,
                self.temp_config.CONTROLS_CONFIG.target_end_year + 1,
            ),
        ]
        benchmark_projection.index = sectors.index

        return benchmark_projection


# FIXME: Need to reshape the tables TARGET_DATA and TRAJECTORY_DATA so scope is a column and the EI data relates only to that scope (wide to long)
class DataVaultWarehouse(DataWarehouse):
    def __init__(
        self,
        engine: sqlalchemy.engine.base.Engine,
        company_data: VaultCompanyDataProvider,
        # This arrives as a table instantiated in the database
        benchmark_projected_production: VaultProviderProductionBenchmark,
        # This arrives as a table instantiated in the database
        benchmarks_projected_ei: VaultProviderIntensityBenchmark,
        ingest_schema: str = None,
        itr_prefix: str = "",
        column_config: Type[ColumnsConfig] = ColumnsConfig,
    ):
        super().__init__(
            company_data=None,
            benchmark_projected_production=None,
            benchmarks_projected_ei=None,
            column_config=column_config,
        )
        self._engine = engine
        self._schema = ingest_schema or engine.dialect.default_schema_name or "demo_dv"
        self._tempscore_table = f"{itr_prefix}temperature_scores"

        # intensity_projections = read_quantified_sql(f"select * from {self._schema}.{self._target_table}", self._target_table, self._schema, self._engine)
        # intensity_projections['scope'] = 'S1S2'
        # intensity_projections['source'] = self._schema

        # If there's no company data, we are just using the vault, not initializing it
        if company_data is None:
            return
        if benchmark_projected_production is None and benchmarks_projected_ei is None:
            return

        # The DataVaultWarehouse provides three calculations per company:
        #    * Cumulative trajectory of emissions
        #    * Cumulative target of emissions
        #    * Cumulative budget of emissions (separately for each benchmark)

        qres = osc._do_sql(
            f"drop table if exists {self._schema}.{itr_prefix}cumulative_emissions",
            self._engine,
            verbose=False,
        )
        qres = osc._do_sql(
            f"""
create table {self._schema}.{itr_prefix}cumulative_emissions with (
    format = 'ORC',
    partitioning = array['scope']
) as
select C.company_name, C.company_id, '{company_data._schema}' as source, 'S1S2' as scope,
       sum((ET.ei_s1_by_year+if(is_nan(ET.ei_s2_by_year),0.0,ET.ei_s2_by_year)) * P.production_by_year) as cumulative_trajectory,
       concat(ET.ei_s1_by_year_units, ' * ', P.production_by_year_units) as cumulative_trajectory_units,
       sum((EI.ei_s1_by_year+if(is_nan(EI.ei_s2_by_year),0.0,EI.ei_s2_by_year)) * P.production_by_year) as cumulative_target,
       concat(EI.ei_s1_by_year_units, ' * ', P.production_by_year_units) as cumulative_target_units
from {company_data._schema}.{company_data._company_table} C
     join {company_data._schema}.{company_data._production_table} P on P.company_id=C.company_id
     join {company_data._schema}.{company_data._target_table} EI on EI.company_id=C.company_id and EI.year=P.year and EI.ei_s1_by_year is not NULL
     join {company_data._schema}.{company_data._trajectory_table} ET on ET.company_id=C.company_id and ET.year=P.year and ET.ei_s1_by_year is not NULL
where P.year>=2020
group by C.company_name, C.company_id, '{company_data._schema}', 'S1S2',
         concat(ET.ei_s1_by_year_units, ' * ', P.production_by_year_units),
         concat(EI.ei_s1_by_year_units, ' * ', P.production_by_year_units)
UNION ALL
select C.company_name, C.company_id, '{company_data._schema}' as source, 'S1S2S3' as scope,
       sum((ET.ei_s1_by_year+if(is_nan(ET.ei_s2_by_year),0.0,ET.ei_s2_by_year)+if(is_nan(ET.ei_s3_by_year),0.0,ET.ei_s3_by_year)) * P.production_by_year) as cumulative_trajectory,
       concat(ET.ei_s1_by_year_units, ' * ', P.production_by_year_units) as cumulative_trajectory_units,
       sum((EI.ei_s1_by_year+if(is_nan(EI.ei_s2_by_year),0.0,EI.ei_s2_by_year)+if(is_nan(EI.ei_s3_by_year),0.0,EI.ei_s3_by_year)) * P.production_by_year) as cumulative_target,
       concat(EI.ei_s1_by_year_units, ' * ', P.production_by_year_units) as cumulative_target_units
from {company_data._schema}.{company_data._company_table} C
     join {company_data._schema}.{company_data._production_table} P on P.company_id=C.company_id
     join {company_data._schema}.{company_data._target_table} EI on EI.company_id=C.company_id and EI.year=P.year and EI.ei_s1_by_year is not NULL
     join {company_data._schema}.{company_data._trajectory_table} ET on ET.company_id=C.company_id and ET.year=P.year and ET.ei_s1_by_year is not NULL
where P.year>=2020
group by C.company_name, C.company_id, '{company_data._schema}', 'S1S2S3',
         concat(ET.ei_s1_by_year_units, ' * ', P.production_by_year_units),
         concat(EI.ei_s1_by_year_units, ' * ', P.production_by_year_units)
""",
            self._engine,
            verbose=True,
        )

        qres = osc._do_sql(
            f"drop table if exists {self._schema}.{itr_prefix}cumulative_budget_1",
            self._engine,
            verbose=False,
        )
        qres = osc._do_sql(
            f"""
create table {self._schema}.{itr_prefix}cumulative_budget_1 with (
    format = 'ORC',
    partitioning = array['scope']
) as
select C.company_name, C.company_id, '{company_data._schema}' as source, B.scope, 'benchmark_1' as benchmark,
       B.global_budget, B.benchmark_temp,
       sum(B.intensity * P.production_by_year) as cumulative_budget,
       concat(B.intensity_units, ' * ', P.production_by_year_units) as cumulative_budget_units
from {company_data._schema}.{company_data._company_table} C
     join {company_data._schema}.{company_data._production_table} P on P.company_id=C.company_id
     join {self._schema}.{benchmarks_projected_ei.benchmark_name} B on P.year=B.year and C.region=B.region and C.sector=B.sector
where P.year>=2020
group by C.company_name, C.company_id, '{company_data._schema}', B.scope, 'benchmark_1', B.global_budget, B.benchmark_temp,
         concat(B.intensity_units, ' * ', P.production_by_year_units)
""",
            self._engine,
            verbose=True,
        )

    def quant_init(
        self,
        engine: sqlalchemy.engine.base.Engine,
        company_data: VaultCompanyDataProvider,
        ingest_schema: str = None,
        itr_prefix: str = "",
    ):
        # The Quant users of the DataVaultWarehouse produces two calculations per company:
        #    * Target and Trajectory overshoot ratios
        #    * Temperature Scores

        qres = osc._do_sql(
            f"drop table if exists {self._schema}.{itr_prefix}overshoot_ratios",
            self._engine,
            verbose=False,
        )
        qres = osc._do_sql(
            f"""
create table {self._schema}.{itr_prefix}overshoot_ratios with (
    format = 'ORC',
    partitioning = array['scope']
) as
select E.company_name, E.company_id, '{company_data._schema}' as source, B.scope, 'benchmark_1' as benchmark,
       B.global_budget, B.benchmark_temp,
       E.cumulative_trajectory/B.cumulative_budget as trajectory_overshoot_ratio,
       concat(E.cumulative_trajectory_units, ' / (', B.cumulative_budget_units, ')') as trajectory_overshoot_ratio_units,
       E.cumulative_target/B.cumulative_budget as target_overshoot_ratio,
       concat(E.cumulative_target_units, ' / (', B.cumulative_budget_units, ')') as target_overshoot_ratio_units
from {self._schema}.{itr_prefix}cumulative_emissions E
     join {self._schema}.{itr_prefix}cumulative_budget_1 B on E.company_id=B.company_id and E.scope=B.scope
""",
            self._engine,
            verbose=True,
        )

        qres = osc._do_sql(
            f"drop table if exists {self._schema}.{self._tempscore_table}",
            self._engine,
            verbose=False,
        )
        qres = osc._do_sql(
            f"""
create table {self._schema}.{self._tempscore_table} with (
    format = 'ORC',
    partitioning = array['scope']
) as
select R.company_name, R.company_id, '{company_data._schema}' as source, R.scope, 'benchmark_1' as benchmark,
       R.benchmark_temp + R.global_budget * (R.trajectory_overshoot_ratio-1) * 2.2/3664.0 as trajectory_temperature_score,
       'delta_degC' as trajectory_temperature_score_units,
       R.benchmark_temp + R.global_budget * (R.target_overshoot_ratio-1) * 2.2/3664.0 as target_temperature_score,
       'delta_degC' as target_temperature_score_units
from {self._schema}.{itr_prefix}overshoot_ratios R
""",
            self._engine,
            verbose=True,
        )

    def get_preprocessed_company_data(self, company_ids: List[str]) -> List[ICompanyAggregates]:
        raise NotImplementedError

    def get_pa_temp_scores(self, probability: float, company_ids: List[str]) -> pd.Series:
        if probability < 0 or probability > 1:
            raise ValueError(f"probability value {probability} outside range [0.0, 1.0]")
        temp_scores = read_quantified_sql(
            f"select company_id, scope, target_temperature_score, target_temperature_score_units, trajectory_temperature_score, trajectory_temperature_score_units from {self._schema}.{self._tempscore_table}",
            self._tempscore_table,
            self._schema,
            self._engine,
            index_col=["company_id", "scope"],
        )
        # We may have company_ids in our portfolio not in our database, and vice-versa.
        # Return proper pa_temp_scores for what we can find, and np.nan for those we cannot
        retval = pd.Series(data=None, index=company_ids, dtype="float64")
        retval.loc[
            retval.index.intersection(temp_scores.index)
        ] = temp_scores.target_temperature_score * probability + temp_scores.trajectory_temperature_score * (
            1 - probability
        )
        return retval



import logging
import warnings  # needed until apply behaves better with Pint quantities in arrays
from abc import ABC
from typing import List, Type

import numpy as np
import pandas as pd
from pint import DimensionalityError
from pydantic import ValidationError

import ITR

from ..configs import ColumnsConfig, LoggingConfig
from ..data import PA_, Q_
from ..data.data_providers import (
    CompanyDataProvider,
    IntensityBenchmarkDataProvider,
    ProductionBenchmarkDataProvider,
)
from ..data.osc_units import EmissionsQuantity, Quantity_type, asPintDataFrame
from ..interfaces import (
    DF_ICompanyEIProjections,
    EScope,
    ICompanyAggregates,
    ICompanyData,
    ICompanyEIProjection,
    ICompanyEIProjections,
    IEIRealization,
    IEmissionRealization,
    IHistoricData,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
LoggingConfig.add_config_to_logger(logger)


class DataWarehouse(ABC):
    """
    General data provider super class.
    """

    def __init__(
        self,
        company_data: CompanyDataProvider,
        benchmark_projected_production: ProductionBenchmarkDataProvider,
        benchmarks_projected_ei: IntensityBenchmarkDataProvider,
        estimate_missing_data=None,
        column_config: Type[ColumnsConfig] = ColumnsConfig,
    ):
        """
        Create a new data warehouse instance.

        :param company_data: CompanyDataProvider
        :param benchmark_projected_production: ProductionBenchmarkDataProvider
        :param benchmarks_projected_ei: IntensityBenchmarkDataProvider
        """
        self.benchmark_projected_production = None
        self.benchmarks_projected_ei = None
        # benchmarks_projected_ei._EI_df_t is the (transposed) EI dataframe for the benchmark
        # benchmark_projected_production.get_company_projected_production(company_sector_region_scope) gives production data per company (per year)
        # multiplying these two gives aligned emissions data for the company, in case we want to add missing data based on sector averages
        self.column_config = column_config
        self.company_data = company_data
        self.estimate_missing_data = estimate_missing_data
        # Place to stash historic data before doing PC-conversion so it can be retreived when switching to non-PC benchmarks
        self.orig_historic_data = {}
        self.company_scope = {}

        # Production benchmark data is needed to project trajectories
        # Trajectories + Emissions Intensities benchmark data are needed to estimate missing S3 data
        # Target projections rely both on Production benchmark data and S3 estimated data
        # Production-centric manipulations must happen after targets have been projected
        if benchmark_projected_production is not None or benchmarks_projected_ei is not None:
            self.update_benchmarks(benchmark_projected_production, benchmarks_projected_ei)

    def _preserve_historic_data(self):
        for c in self.company_data._companies:
            self.orig_historic_data[c.company_id] = {
                "ghg_s1s2": c.ghg_s1s2,
                "ghg_s3": c.ghg_s3,
                "historic_data": IHistoricData(
                    productions=c.historic_data.productions,
                    emissions=c.historic_data.emissions.model_copy(),
                    emissions_intensities=c.historic_data.emissions_intensities.model_copy(),
                ),
                "projected_intensities": c.projected_intensities.model_copy(),
                "projected_targets": c.projected_targets.model_copy(),
            }
            orig_historic_data = self.orig_historic_data[c.company_id]
            for scope_name in EScope.get_scopes():
                intensity_scope = getattr(orig_historic_data["projected_intensities"], scope_name, None)
                if intensity_scope is not None:
                    setattr(
                        orig_historic_data["projected_intensities"],
                        scope_name,
                        (
                            type(intensity_scope)(
                                ei_metric=intensity_scope.ei_metric,
                                projections=intensity_scope.projections.copy(),
                            )
                        ),
                    )
                target_scope = getattr(orig_historic_data["projected_targets"], scope_name, None)
                if target_scope is not None:
                    setattr(
                        orig_historic_data["projected_targets"],
                        scope_name,
                        (
                            type(target_scope)(
                                ei_metric=target_scope.ei_metric,
                                projections=target_scope.projections.copy(),
                            )
                        ),
                    )

    def _restore_historic_data(self):
        for c in self.company_data._companies:
            orig_data = self.orig_historic_data[c.company_id]
            c.ghg_s1s2 = orig_data["ghg_s1s2"]
            c.ghg_s3 = orig_data["ghg_s3"]
            c.historic_data = orig_data["historic_data"]
            c.projected_intensities = orig_data["projected_intensities"]
            c.projected_targets = orig_data["projected_targets"]

    def update_benchmarks(
        self,
        benchmark_projected_production: ProductionBenchmarkDataProvider,
        benchmarks_projected_ei: IntensityBenchmarkDataProvider,
    ):
        """
        Update the benchmark data used in this instance of the DataWarehouse.  If there is no change, do nothing.
        """
        new_production_bm = new_ei_bm = new_prod_centric = False
        assert benchmark_projected_production is not None
        if (
            self.benchmark_projected_production is None
            or self.benchmark_projected_production._productions_benchmarks
            != benchmark_projected_production._productions_benchmarks
        ):
            self.benchmark_projected_production = benchmark_projected_production
            new_production_bm = True
        assert benchmarks_projected_ei is not None
        if (
            self.benchmarks_projected_ei is None
            or self.benchmarks_projected_ei._EI_benchmarks != benchmarks_projected_ei._EI_benchmarks
        ):
            prev_prod_centric = next_prod_centric = False
            if self.benchmarks_projected_ei is not None and getattr(
                self.benchmarks_projected_ei._EI_benchmarks, "S1S2", None
            ):
                prev_prod_centric = self.benchmarks_projected_ei._EI_benchmarks["S1S2"].production_centric
            if getattr(benchmarks_projected_ei._EI_benchmarks, "S1S2", None):
                next_prod_centric = benchmarks_projected_ei._EI_benchmarks["S1S2"].production_centric
            new_prod_centric = prev_prod_centric != next_prod_centric
            self.benchmarks_projected_ei = benchmarks_projected_ei
            new_ei_bm = True

        # Production benchmark data is needed to project trajectories
        # Trajectories + Emissions Intensities benchmark data are needed to estimate missing S3 data
        # Target projections rely both on Production benchmark data and S3 estimated data
        # Production-centric benchmarks shift S3 data after trajectory and targets have been projected
        if new_production_bm:
            logger.info(
                f"new_production_bm calculating trajectories for {len(self.company_data._companies)} companies (times {len(EScope.get_scopes())} scopes times {self.company_data.projection_controls.TARGET_YEAR-self.company_data.projection_controls.BASE_YEAR} years)"
            )
            self.company_data._validate_projected_trajectories(
                self.company_data._companies, self.benchmarks_projected_ei._EI_df_t
            )

        if new_ei_bm and (new_companies := [c for c in self.company_data._companies if "+" in c.company_id]):
            logger.info("Allocating emissions to align with benchmark data")
            bm_prod_df = benchmark_projected_production._prod_df
            bm_ei_df_t = benchmarks_projected_ei._EI_df_t
            bm_sectors = bm_ei_df_t.columns.get_level_values("sector").unique().to_list()
            base_year = self.company_data.projection_controls.BASE_YEAR

            from collections import defaultdict

            sectors_dict = defaultdict(list)
            region_dict = {}
            historic_dict = {}

            for c in new_companies:
                orig_id, sector = c.company_id.split("+")
                if sector in bm_sectors:
                    sectors_dict[orig_id].append(sector)
                else:
                    logger.error(f"No benchmark sector data for {orig_id}: sector = {sector}")
                    continue
                if (sector, c.region) in bm_ei_df_t.columns:
                    region_dict[orig_id] = c.region
                elif (sector, "Global") in bm_ei_df_t.columns:
                    region_dict[orig_id] = "Global"
                else:
                    logger.error(f"No benchmark region data for {orig_id}: sector = {sector}; region = {region}")
                    continue

                # Though we mutate below, it's our own unique copy of c.historic_data we are mutating, so OK
                historic_dict[c.company_id] = c.historic_data

            for orig_id, sectors in sectors_dict.items():
                region = region_dict[orig_id]
                sector_ei = [
                    (
                        sector,
                        scope,
                        # bm_prod_df.loc[(sector, region, EScope.AnyScope)][base_year] is '1.0 dimensionless'
                        bm_ei_df_t.loc[:, (sector, region, scope)][base_year],
                    )
                    for scope in EScope.get_result_scopes()
                    # This only saves us from having data about sectorized alignments we might not need.  It doesn't affect the emissions being allocated (or not).
                    if (c.company_id, scope.name) in self.company_data._bm_allocation_index
                    for sector in sectors
                    if (sector, region, scope) in bm_ei_df_t.columns
                    and historic_dict["+".join([orig_id, sector])].emissions[scope.name]
                ]
                sector_ei_df = pd.DataFrame(sector_ei, columns=["sector", "scope", "ei"]).set_index(["sector"])
                sector_prod_df = pd.DataFrame(
                    [
                        (sector, prod.value)
                        for sector in sectors
                        for prod in historic_dict["+".join([orig_id, sector])].productions
                        # FIXME: if we don't have proudction values for BASE_YEAR, this fails!  See 'US2333311072+Gas Utilities'
                        if prod.year == base_year
                    ],
                    columns=["sector", "prod"],
                ).set_index("sector")
                sector_em_df = (
                    sector_ei_df.join(sector_prod_df)
                    .assign(em=lambda x: x["ei"] * x["prod"])
                    .set_index("scope", append=True)
                    .drop(columns=["ei", "prod"])
                )
                # Only now can we drop whatever is not in self.company_data._bm_allocation_index from sector_em_df
                if self.company_data._bm_allocation_index.empty:
                    # No allocations to make for any companies
                    continue
                to_allocate_idx = self.company_data._bm_allocation_index[
                    self.company_data._bm_allocation_index.map(lambda x: x[0].startswith(orig_id))
                ].map(lambda x: (x[0].split("+")[1], EScope[x[1]]))
                if to_allocate_idx.empty:
                    logger.info(f"Already allocated emissions for {orig_id} across {sectors}")
                    continue
                # FIXME: to_allocate_idx is missing S1S2S3 for US2091151041
                to_allocate_idx.names = ["sector", "scope"]
                try:
                    sector_em_df = sector_em_df.loc[sector_em_df.index.intersection(to_allocate_idx)].astype(
                        "pint[Mt CO2e]"
                    )
                except DimensionalityError:
                    # breakpoint()
                    assert False
                em_tot = sector_em_df.groupby("scope")["em"].sum()
                # The alignment calculation: Company Scope-Sector emissions = Total Company Scope emissions * (BM Scope Sector / SUM(All Scope Sectors of Company))
                aligned_em = [
                    (
                        sector,
                        [
                            (
                                scope,
                                list(
                                    map(
                                        lambda em: (
                                            em[0],
                                            em[1]
                                            * sector_em_df.loc[(sector, scope)].squeeze()
                                            / em_tot.loc[scope].squeeze(),
                                        ),
                                        [
                                            (em.year, em.value)
                                            for em in historic_dict["+".join([orig_id, sector])].emissions[scope.name]
                                        ],
                                    )
                                ),
                            )
                            for scope in em_tot.index
                            if em_tot.loc[scope].squeeze().m != 0.0
                        ],
                    )
                    for sector in sectors
                ]

                # Having done all scopes and sectors for this company above, replace historic Em and EI data below
                for sector_aligned in aligned_em:
                    sector, scopes = sector_aligned
                    historic_sector = historic_dict["+".join([orig_id, sector])]
                    # print(f"Historic {sector} initially\n{historic_sector.emissions}")
                    for scope_tuple in scopes:
                        scope, em_list = scope_tuple
                        setattr(
                            historic_sector.emissions,
                            scope.name,
                            list(
                                map(
                                    lambda em: IEmissionRealization(year=em[0], value=em[1].to("Mt CO2e")),
                                    em_list,
                                )
                            ),
                        )
                        prod_list = historic_sector.productions
                        ei_list = list(
                            map(
                                lambda em_p: IEIRealization(
                                    year=em_p[0].year,
                                    value=Q_(
                                        np.nan,
                                        f"({em_p[0].value.u}) / ({em_p[1].value.u})",
                                    )
                                    if em_p[1].value.m == 0.0
                                    else em_p[0].value / em_p[1].value,
                                ),
                                zip(historic_sector.emissions[scope.name], prod_list),
                            )
                        )
                        setattr(historic_sector.emissions_intensities, scope.name, ei_list)
                    # print(f"Historic {sector} adjusted\n{historic_dict['+'.join([orig_id, sector])].emissions}")
            logger.info("Sector alignment complete")

        # If we are missing S3 (or other) data, fill in before projecting targets
        if new_ei_bm and self.estimate_missing_data is not None:
            logger.info(f"estimating missing data")
            for c in self.company_data._companies:
                self.estimate_missing_data(self, c)

        # Changes to production benchmark requires re-calculating targets (which are production-dependent)
        if new_production_bm:
            logger.info(
                f"projecting targets for {len(self.company_data._companies)} companies (times {len(EScope.get_scopes())} scopes times {self.company_data.projection_controls.TARGET_YEAR-self.company_data.projection_controls.BASE_YEAR} years)"
            )
            self.company_data._calculate_target_projections(benchmark_projected_production, benchmarks_projected_ei)

        # If our benchmark is production-centric, migrate S3 data (including estimated S3 data) into S1S2
        # If we shift before we project, then S3 targets will not be projected correctly.
        if (
            new_ei_bm
            and getattr(benchmarks_projected_ei._EI_benchmarks, "S1S2", None)
            and benchmarks_projected_ei._EI_benchmarks["S1S2"].production_centric
        ):
            logger.info(f"Shifting S3 emissions data into S1 according to Production-Centric benchmark rules")
            if self.orig_historic_data != {}:
                self._restore_historic_data()
            else:
                self._preserve_historic_data()
            for c in self.company_data._companies:
                if c.ghg_s3:
                    # For Production-centric and energy-only data (except for Cement), convert all S3 numbers to S1 numbers
                    if not ITR.isna(c.ghg_s3):
                        c.ghg_s1s2 = c.ghg_s1s2 + c.ghg_s3
                    c.ghg_s3 = None  # Q_(0.0, c.ghg_s3.u)
                if c.historic_data:

                    def _adjust_historic_data(data, primary_scope_attr, data_adder):
                        if data[primary_scope_attr]:
                            pre_s3_data = [p for p in data[primary_scope_attr] if p.year <= data.S3[0].year]
                            if len(pre_s3_data) == 0:
                                # Could not adjust
                                # breakpoint()
                                assert False
                                return
                            if len(pre_s3_data) > 1:
                                pre_s3_data = list(
                                    map(
                                        lambda x: type(x)(
                                            year=x.year,
                                            value=data.S3[0].value * x.value / pre_s3_data[-1].value,
                                        ),
                                        pre_s3_data[:-1],
                                    )
                                )
                                s3_data = pre_s3_data + data.S3
                            else:
                                s3_data = data.S3
                            setattr(
                                data,
                                primary_scope_attr,
                                list(map(data_adder, data[primary_scope_attr], s3_data)),
                            )
                        else:
                            setattr(data, primary_scope_attr, data.S3)

                    if c.historic_data.emissions and c.historic_data.emissions.S3:
                        _adjust_historic_data(c.historic_data.emissions, "S1", IEmissionRealization.add)
                        _adjust_historic_data(c.historic_data.emissions, "S1S2", IEmissionRealization.add)
                        c.historic_data.emissions.S3 = []
                    if c.historic_data.emissions and c.historic_data.emissions.S1S2S3:
                        # assert c.historic_data.emissions.S1S2 == c.historic_data.emissions.S1S2S3
                        c.historic_data.emissions.S1S2S3 = []
                    if c.historic_data.emissions_intensities and c.historic_data.emissions_intensities.S3:
                        _adjust_historic_data(
                            c.historic_data.emissions_intensities,
                            "S1",
                            IEIRealization.add,
                        )
                        _adjust_historic_data(
                            c.historic_data.emissions_intensities,
                            "S1S2",
                            IEIRealization.add,
                        )
                        c.historic_data.emissions_intensities.S3 = []
                    if c.historic_data.emissions_intensities and c.historic_data.emissions_intensities.S1S2S3:
                        # assert c.historic_data.emissions_intensities.S1S2 == c.historic_data.emissions_intensities.S1S2S3
                        c.historic_data.emissions_intensities.S1S2S3 = []
                if c.projected_intensities and c.projected_intensities.S3:

                    def _adjust_trajectories(trajectories, primary_scope_attr):
                        if not trajectories[primary_scope_attr]:
                            setattr(trajectories, primary_scope_attr, trajectories.S3)
                        else:
                            if isinstance(trajectories.S3.projections, pd.Series):
                                trajectories[primary_scope_attr].projections = trajectories[
                                    primary_scope_attr
                                ].projections.add(trajectories.S3.projections)
                            else:
                                # Should not be reached as we are using DF_ICompanyEIProjections consistently now
                                # breakpoint()
                                assert False
                                trajectories[primary_scope_attr].projections = list(
                                    map(
                                        ICompanyEIProjection.add,
                                        trajectories[primary_scope_attr].projections,
                                        trajectories.S3.projections,
                                    )
                                )

                    _adjust_trajectories(c.projected_intensities, "S1")
                    _adjust_trajectories(c.projected_intensities, "S1S2")
                    c.projected_intensities.S3 = None
                    if c.projected_intensities.S1S2S3:
                        # assert c.projected_intensities.S1S2.projections == c.projected_intensities.S1S2S3.projections
                        c.projected_intensities.S1S2S3 = None
                if c.projected_targets and c.projected_targets.S3:
                    # For production-centric benchmarks, S3 emissions are counted against S1 (and/or the S1 in S1+S2)
                    def _align_and_sum_projected_targets(targets, primary_scope_attr):
                        if targets[primary_scope_attr] is None:
                            raise AttributeError
                        primary_projections = targets[primary_scope_attr].projections
                        s3_projections = targets.S3.projections
                        if isinstance(s3_projections, pd.Series):
                            targets[primary_scope_attr].projections = (
                                # We should convert S3 data from benchmark-type to disclosed-type earlier in the chain
                                primary_projections
                                + s3_projections.astype(primary_projections.dtype)
                            )
                        else:
                            # Should not be reached as we are using DF_ICompanyEIProjections consistently now
                            # breakpoint()
                            assert False
                            if primary_projections[0].year < s3_projections[0].year:
                                while primary_projections[0].year < s3_projections[0].year:
                                    primary_projections = primary_projections[1:]
                            elif primary_projections[0].year > s3_projections[0].year:
                                while primary_projections[0].year > s3_projections[0].year:
                                    s3_projections = s3_projections[1:]
                            targets[primary_scope_attr].projections = list(
                                map(
                                    ICompanyEIProjection.add,
                                    primary_projections,
                                    s3_projections,
                                )
                            )

                    if c.projected_targets.S1:
                        _align_and_sum_projected_targets(c.projected_targets, "S1")
                    try:
                        # S3 projected targets may have been synthesized from a netzero S1S2S3 target and might need to be date-aligned with S1S2
                        _align_and_sum_projected_targets(c.projected_targets, "S1S2")
                    except AttributeError:
                        if c.projected_targets.S2:
                            logger.warning(
                                f"Scope 1+2 target projections should have been created for {c.company_id}; repairing"
                            )
                            if isinstance(c.projected_targets.S1.projections, pd.Series) and isinstance(
                                c.projected_targets.S2.projections, pd.Series
                            ):
                                c.projected_targets.S1S2 = DF_ICompanyEIProjections(
                                    ei_metric=c.projected_targets.S1.ei_metric,
                                    projections=c.projected_targets.S1.projections + c.projected_targets.S2.projections,
                                )
                            else:
                                c.projected_targets.S1S2 = ICompanyEIProjections(
                                    ei_metric=c.projected_targets.S1.ei_metric,
                                    projections=list(
                                        map(
                                            ICompanyEIProjection.add,
                                            c.projected_targets.S1.projections,
                                            c.projected_targets.S2.projections,
                                        )
                                    ),
                                )
                        else:
                            logger.warning(
                                f"Scope 2 target projections missing from company with ID {c.company_id}; treating as zero"
                            )
                            if isinstance(c.projected_targets.S1.projections, pd.Series):
                                c.projected_targets.S1S2 = DF_ICompanyEIProjections(
                                    ei_metric=c.projected_targets.S1.ei_metric,
                                    projections=c.projected_targets.S1.projections,
                                )
                            else:
                                c.projected_targets.S1S2 = ICompanyEIProjections(
                                    ei_metric=c.projected_targets.S1.ei_metric,
                                    projections=c.projected_targets.S1.projections,
                                )
                        if c.projected_targets.S3:
                            _align_and_sum_projected_targets(c.projected_targets, "S1S2")
                        else:
                            logger.warning(
                                f"Scope 3 target projections missing from company with ID {c.company_id}; treating as zero"
                            )
                    except ValueError:
                        logger.error(
                            f"S1+S2 targets not aligned with S3 targets for company with ID {c.company_id}; ignoring S3 data"
                        )
                    c.projected_targets.S3 = None
                if c.projected_targets and c.projected_targets.S1S2S3:
                    # assert c.projected_targets.S1S2 == c.projected_targets.S1S2S3
                    c.projected_targets.S1S2S3 = None
        elif new_prod_centric and self.orig_historic_data != {}:
            # Switch to non-product-centric benchmark of this version.
            self._restore_historic_data()

        # Set scope information based on what company reports and what benchmark requres
        # benchmarks_projected_ei._EI_df_t makes life a bit easier...
        missing_company_scopes = []
        for c in self.company_data._companies:
            region = c.region
            try:
                bm_company_sector_region = benchmarks_projected_ei._EI_df_t[(c.sector, region)]
            except KeyError:
                try:
                    region = "Global"
                    bm_company_sector_region = benchmarks_projected_ei._EI_df_t[(c.sector, region)]
                except KeyError:
                    missing_company_scopes.append(c.company_id)
                    continue
            scopes = benchmarks_projected_ei._EI_df_t[(c.sector, region)].columns.tolist()
            if len(scopes) == 1:
                self.company_scope[c.company_id] = scopes[0]
                continue
            for scope in EScope.get_result_scopes():
                if scope in scopes:
                    self.company_scope[c.company_id] = scope
                    break
            if c.company_id not in self.company_scope:
                missing_company_scopes.append(c.company_id)

        if missing_company_scopes:
            logger.warning(
                f"The following companies do not disclose scope data required by benchmark and will be not be analyzed: {missing_company_scopes}"
            )

    def update_trajectories(self):
        """
        Update the trajectory calculations after changing global ProjectionControls.  Production and EI benchmarks remain the same.
        """
        # We cannot only update trajectories without regard for all that depend on those trajectories
        # For example, different benchmarks may have different scopes defined, units for benchmarks, etc.
        logger.info(
            f"re-calculating trajectories for {len(self.company_data._companies)} companies\n    (times {len(EScope.get_scopes())} scopes times {self.company_data.projection_controls.TARGET_YEAR-self.company_data.projection_controls.BASE_YEAR} years)"
        )
        for company in self.company_data._companies:
            company.projected_intensities = None
        self.company_data._validate_projected_trajectories(
            self.company_data._companies, self.benchmarks_projected_ei._EI_df_t
        )

    def estimate_missing_s3_data(self, company: ICompanyData) -> None:
        # We need benchmark data to estimate S3 from projected_intensities (which go back in time to BASE_YEAR).
        # We don't need to estimate further back than that, as we don't rewrite values stored in historic_data.
        # Benchmarks that don't have S3 emissions (such as TPI for Electricity Utilities) take this early return
        if (
            company.projected_intensities.S3
            or not self.benchmarks_projected_ei
            or not self.benchmarks_projected_ei._EI_benchmarks.S3
        ):
            return

        # This is an estimation function, not a mathematical processing function.
        # It won't solve S3 = S1S2S3 - S1S2 (which should be done elsewhere)
        sector = company.sector
        region = company.region
        if (sector, region) in self.benchmarks_projected_ei._EI_df_t.columns:
            pass
        elif (sector, "Global") in self.benchmarks_projected_ei._EI_df_t.columns:
            region = "Global"
        else:
            return
        if (
            sector,
            region,
            EScope.S3,
        ) not in self.benchmarks_projected_ei._EI_df_t.columns:
            if sector == "Construction Buildings":
                # Construction Buildings don't have an S3 scope defined
                return
            else:
                # Some benchmarks don't include S3 for all sectors; so nothing to estimate
                return
        else:
            bm_ei_s3 = self.benchmarks_projected_ei._EI_df_t[(sector, region, EScope.S3)]
        if (
            sector,
            region,
            EScope.S1S2,
        ) in self.benchmarks_projected_ei._EI_df_t.columns:
            # If we have only S1S2S3 emissions, we can "allocate" them according to the benchmark's allocation
            bm_ei_s1s2 = self.benchmarks_projected_ei._EI_df_t[(sector, region, EScope.S1S2)]
        else:
            bm_ei_s1s2 = None

        if company.projected_intensities.S1S2S3:
            assert company.projected_intensities.S1S2 is None
            assert company.projected_intensities.S1 is None
            ei_metric = company.projected_intensities.S1S2S3.ei_metric
            s1s2_projections = company.projected_intensities.S1S2S3.projections * bm_ei_s1s2 / (bm_ei_s1s2 + bm_ei_s3)
            s3_projections = company.projected_intensities.S1S2S3.projections * bm_ei_s3 / (bm_ei_s1s2 + bm_ei_s3)
            if ITR.HAS_UNCERTAINTIES:
                nominal_s3 = ITR.nominal_values(s3_projections.pint.quantity.m)
                std_dev_s3 = ITR.uarray(np.zeros(len(nominal_s3)), nominal_s3)
                s1s2_projections = pd.Series(
                    data=PA_(
                        s1s2_projections.pint.quantity.m + std_dev_s3,
                        dtype=s1s2_projections.dtype,
                    ),
                    index=s1s2_projections.index,
                )
                s3_projections = pd.Series(
                    data=PA_(
                        s3_projections.pint.quantity.m + std_dev_s3,
                        dtype=s3_projections.dtype,
                    ),
                    index=s3_projections.index,
                )
            company.projected_intensities.S1S2 = DF_ICompanyEIProjections(
                ei_metric=ei_metric, projections=s1s2_projections
            )
            company.projected_intensities.S3 = DF_ICompanyEIProjections(ei_metric=ei_metric, projections=s3_projections)
            company.ghg_s1s2 = (
                s1s2_projections[self.company_data.projection_controls.BASE_YEAR] * company.base_year_production
            ).to("t CO2e")
        else:
            # Penalize non-disclosure by assuming 2x aligned S3 emissions.  That's still likely undercounting, because
            # most non-disclosing companies are nowhere near the reduction rates of the benchmarks.
            # It would be lovely to use S1S2 or S1 data to inform S3, but that likely adds error on top of error
            assert company.projected_intensities.S1S2S3 is None
            ei_metric = str(bm_ei_s3.dtype.units)
            # If we don't have uncertainties, ITR.ufloat just returns the nom value 2.0
            s3_projections = bm_ei_s3 * ITR.ufloat(2.0, 1.0)
            company.projected_intensities.S3 = DF_ICompanyEIProjections(ei_metric=ei_metric, projections=s3_projections)
            if company.projected_intensities.S1S2 is not None:
                try:
                    company.projected_intensities.S1S2S3 = DF_ICompanyEIProjections(
                        ei_metric=ei_metric,
                        # Mismatched dimensionalities are not automatically converted
                        projections=s3_projections
                        + company.projected_intensities.S1S2.projections.astype(f"pint[{ei_metric}]"),
                    )
                except DimensionalityError:
                    logger.error(
                        f"Company {company.company_id}'s S1+S2 intensity units "
                        f"({company.projected_intensities.S1S2.projections.dtype})"
                        f"are not compatible with benchmark units ({ei_metric})"
                    )

        company.ghg_s3 = (
            s3_projections[self.company_data.projection_controls.BASE_YEAR] * company.base_year_production
        ).to("t CO2e")
        logger.info(f"Added S3 estimates for {company.company_id} (sector = {sector}, region = {region})")

    @classmethod
    def _process_company_data(
        cls,
        df_company_data: pd.DataFrame,
        projected_production: pd.DataFrame,
        projected_trajectories: pd.DataFrame,
        projected_targets: pd.DataFrame,
        base_year: int,
        target_year: int,
        budgeted_ei: pd.DataFrame,
        benchmark_temperature: Quantity_type("delta_degC"),
        global_budget: EmissionsQuantity,
    ) -> pd.DataFrame:
        # trajectories are projected from historic data and we are careful to fill all gaps between historic and projections
        # FIXME: we just computed ALL company data above into a dataframe.  Why not use that?  Answer: because the following is inscrutible
        # (lambda xx: pd.DataFrame(data=list(xx.map(lambda x: { 'scope': x.name } | x.to_dict())), index=xx.index)) \
        #     (df_company_data.projected_intensities
        #      .map(lambda x: [pd.Series(x[scope]['projections'], name=getattr(EScope, scope))
        #                      for scope in ['S1', 'S1S2', 'S3', 'S1S2S3']
        #                      if x[scope] is not None]).explode()).set_index('scope', append=True)

        # If we have excess projections (compared to projected_production), _get_cumulative_emissions will drop them
        df_trajectory = DataWarehouse._get_cumulative_emissions(
            projected_ei=projected_trajectories,
            projected_production=projected_production,
        )

        # Ensure we haven't set any targets for scopes we are not prepared to deal with
        projected_targets = projected_targets.loc[projected_production.index.intersection(projected_targets.index)]
        # Fill in ragged left edge of projected_targets with historic data, interpolating where we need to
        for year_col, company_ei_data in projected_targets.items():
            # company_ei_data is an unruly collection of unit types, so need to check NaN values row by row
            mask = company_ei_data.apply(lambda x: ITR.isna(x))
            if mask.all():
                # No sense trying to do anything with left-side all-NaN columns
                projected_targets = projected_targets.drop(columns=year_col)
                continue
            if mask.any():
                projected_targets.loc[mask[mask].index, year_col] = projected_trajectories.loc[
                    mask[mask].index, year_col
                ]
            else:
                break

        df_target = DataWarehouse._get_cumulative_emissions(
            projected_ei=projected_targets, projected_production=projected_production
        )
        df_budget = DataWarehouse._get_cumulative_emissions(
            projected_ei=budgeted_ei, projected_production=projected_production
        )
        base_year_scale = df_trajectory.loc[df_budget.index][base_year].mul(
            df_budget[base_year].map(lambda x: Q_(0.0, f"1/({x.u})") if x.m == 0.0 else 1 / x)
        )
        df_scaled_budget = df_budget.mul(base_year_scale, axis=0)
        # FIXME: we calculate exceedance only against df_budget, not also df_scaled_budget
        # df_trajectory_exceedance = self._get_exceedance_year(df_trajectory, df_budget, self.company_data.projection_controls.TARGET_YEAR, None)
        # df_target_exceedance = self._get_exceedance_year(df_target, df_budget, self.company_data.projection_controls.TARGET_YEAR, None)
        df_trajectory_exceedance = DataWarehouse._get_exceedance_year(
            df_trajectory, df_budget, base_year, target_year, target_year
        )
        df_target_exceedance = DataWarehouse._get_exceedance_year(
            df_target, df_budget, base_year, target_year, target_year
        )
        df_scope_data = pd.concat(
            [
                df_trajectory.iloc[:, -1].rename(ColumnsConfig.CUMULATIVE_TRAJECTORY),
                df_target.iloc[:, -1].rename(ColumnsConfig.CUMULATIVE_TARGET),
                df_budget.iloc[:, -1].rename(ColumnsConfig.CUMULATIVE_BUDGET),
                df_scaled_budget.iloc[:, -1].rename(ColumnsConfig.CUMULATIVE_SCALED_BUDGET),
                df_trajectory_exceedance.rename(f"{ColumnsConfig.TRAJECTORY_EXCEEDANCE_YEAR}"),
                df_target_exceedance.rename(f"{ColumnsConfig.TARGET_EXCEEDANCE_YEAR}"),
            ],
            axis=1,
        )
        df_company_data = df_company_data.join(df_scope_data).reset_index("scope")
        na_company_mask = df_company_data.scope.isna()
        if na_company_mask.any():
            # Happens when the benchmark doesn't cover the company's supplied scopes at all
            logger.warning(
                f"Dropping companies with no scope data: {df_company_scope[na_company_mask].index.get_level_values(level='company_id').to_list()}"
            )
            df_company_data = df_company_data[~na_company_mask]
        df_company_data[ColumnsConfig.BENCHMARK_GLOBAL_BUDGET] = pd.Series(
            PA_([global_budget.m] * len(df_company_data), dtype=str(global_budget.u)),
            index=df_company_data.index,
        )
        # ICompanyAggregates wants this Quantity as a `str`
        df_company_data[ColumnsConfig.BENCHMARK_TEMP] = [str(benchmark_temperature)] * len(df_company_data)
        return df_company_data

    def get_preprocessed_company_data(self, company_ids: List[str]) -> List[ICompanyAggregates]:
        """
        Get all relevant data for a list of company ids. This method should return a list of ICompanyAggregates
        instances.

        :param company_ids: A list of company IDs (ISINs)
        :return: A list containing the company data and additional precalculated fields
        """

        company_data = self.company_data.get_company_data(company_ids)
        df_company_data = pd.DataFrame.from_records([dict(c) for c in company_data]).set_index(
            self.column_config.COMPANY_ID, drop=False
        )
        valid_company_ids = df_company_data.index.to_list()

        # Why does the following create ICompanyData?  Because get_company_fundamentals is getting base_year_production, ghg_s1s2, and ghg_s3 the hard way
        company_info_at_base_year = self.company_data.get_company_intensity_and_production_at_base_year(
            valid_company_ids
        )

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            # See https://github.com/hgrecco/pint-pandas/issues/128
            projected_production = self.benchmark_projected_production.get_company_projected_production(
                company_info_at_base_year
            )
            target_year_loc = projected_production.columns.get_loc(self.company_data.projection_controls.TARGET_YEAR)
            projected_production = projected_production.iloc[:, 0 : target_year_loc + 1]

        df_company_data = self._process_company_data(
            df_company_data,
            projected_production=projected_production,
            projected_trajectories=self.company_data.get_company_projected_trajectories(valid_company_ids),
            projected_targets=self.company_data.get_company_projected_targets(valid_company_ids),
            base_year=self.company_data.projection_controls.BASE_YEAR,
            target_year=self.company_data.projection_controls.TARGET_YEAR,
            budgeted_ei=self.benchmarks_projected_ei.get_SDA_intensity_benchmarks(company_info_at_base_year),
            benchmark_temperature=self.benchmarks_projected_ei.benchmark_temperature,
            global_budget=self.benchmarks_projected_ei.benchmark_global_budget,
        )

        companies = df_company_data.to_dict(orient="records")
        # This was WICKED SLOW: aggregate_company_data = [ICompanyAggregates.parse_obj(company) for company in companies]
        aggregate_company_data = [
            ICompanyAggregates.from_ICompanyData(company, scope_company_data)
            for company in company_data
            for scope_company_data in df_company_data.loc[[company.company_id]][
                [
                    "cumulative_budget",
                    "cumulative_scaled_budget",
                    "cumulative_trajectory",
                    "cumulative_target",
                    "benchmark_temperature",
                    "benchmark_global_budget",
                    "scope",
                    "trajectory_exceedance_year",
                    "target_exceedance_year",
                ]
            ].to_dict(orient="records")
        ]
        return aggregate_company_data

    def _convert_df_to_model(self, df_company_data: pd.DataFrame) -> List[ICompanyAggregates]:
        """
        transforms Dataframe Company data and preprocessed values into list of ICompanyAggregates instances

        :param df_company_data: pandas Dataframe with targets
        :return: A list containing the targets
        """
        df_company_data = df_company_data.where(pd.notnull(df_company_data), None).replace(
            {np.nan: None}
        )  # set NaN to None since NaN is float instance
        companies_data_dict = df_company_data.to_dict(orient="records")
        model_companies: List[ICompanyAggregates] = []
        for company_data in companies_data_dict:
            try:
                model_companies.append(ICompanyAggregates.model_validate(company_data))
            except ValidationError:
                logger.warning(
                    "(one of) the input(s) of company %s is invalid and will be skipped"
                    % company_data[self.column_config.COMPANY_NAME]
                )
                pass
        return model_companies

    @classmethod
    def _get_cumulative_emissions(cls, projected_ei: pd.DataFrame, projected_production: pd.DataFrame) -> pd.DataFrame:
        """
        get the weighted sum of the projected emission
        :param projected_ei: Rows of projected emissions intensities indexed by (company_id, scope)
        :param projected_production: Rows of projected production amounts indexed by (company_id, scope)
        :return: cumulative emissions, by year, based on weighted sum of emissions intensity * production
        """

        # The old and slow way:
        # projected_ei_t = asPintDataFrame(projected_ei.T)
        # projected_prod_t = asPintDataFrame(projected_production.T)
        # projected_emissions_t = projected_ei_t.mul(projected_prod_t.loc[projected_ei_t.index, projected_ei.T.columns])
        # cumulative_emissions = projected_emissions_t.T.cumsum(axis=1).astype('pint[Mt CO2]')

        # Ensure that projected_production is ordered the same as projected_ei, preserving order of projected_ei
        # projected_production is constructed to be limited to the years we want to analyze
        proj_prod_t = asPintDataFrame(projected_production.loc[projected_ei.index].T)
        # Limit projected_ei to the year range of projected_production
        proj_ei_t = asPintDataFrame(projected_ei[proj_prod_t.index].T)
        units_CO2e = "t CO2e"
        # We use pd.concat because pd.combine reverts our PintArrays into object arrays :-/
        proj_CO2e_m_t = pd.concat(
            [
                ITR.data.osc_units.align_production_to_bm(proj_prod_t[col], proj_ei_t[col])
                .mul(proj_ei_t[col])
                .pint.m_as(units_CO2e)
                for col in proj_ei_t.columns
            ],
            axis=1,
        )
        # pd.concat names parameter refers to index.names; there's no way to set columns.names
        proj_CO2e_m_t.columns.names = proj_ei_t.columns.names
        if ITR.HAS_UNCERTAINTIES:
            # Sum both the nominal and std_dev values, because these series are completely correlated
            # Note that NaNs in this dataframe will be nan+/-nan, showing up in both nom and err
            nom_CO2e_m_t = proj_CO2e_m_t.apply(ITR.nominal_values).cumsum()
            err_CO2e_m_t = proj_CO2e_m_t.apply(ITR.std_devs).cumsum()
            cumulative_emissions_m_t = nom_CO2e_m_t.combine(err_CO2e_m_t, ITR.recombine_nom_and_std)
        else:
            cumulative_emissions_m_t = proj_CO2e_m_t.cumsum()
        return cumulative_emissions_m_t.T.astype(f"pint[{units_CO2e}]")

    @classmethod
    def _get_exceedance_year(
        self,
        df_subject: pd.DataFrame,
        df_budget: pd.DataFrame,
        base_year: int,
        target_year: int,
        budget_year: int = None,
    ) -> pd.Series:
        """
        :param df_subject: DataFrame of cumulative emissions values over time
        :param df_budget: DataFrame of cumulative emissions budget allowed over time
        :param budget_year: if not None, set the exceedence budget to that year; otherwise budget starts low and grows year-by-year
        :return: The furthest-out year where df_subject < df_budget, or np.nan if none

        Where the (df_subject-aligned) budget defines a value but df_subject doesn't have a value, return pd.NA
        Where the benchmark (df_budget) fails to provide a metric for the subject scope, return no rows
        """
        missing_subjects = df_budget.index.difference(df_subject.index)
        aligned_rows = df_budget.index.intersection(df_subject.index)
        # idxmax returns the first maximum of a series, but we want the last maximum of a series
        # Reversing the columns, the maximum remains the maximum, but the "first" is the furthest-out year

        df_subject = df_subject.loc[aligned_rows, ::-1].pint.dequantify()
        df_budget = df_budget.loc[aligned_rows, ::-1].pint.dequantify()
        # units are embedded in the column multi-index, so this check validates dequantify operation post-hoc
        assert (df_subject.columns == df_budget.columns).all()
        if budget_year:
            df_exceedance_budget = pd.DataFrame(
                {
                    (year, units): df_budget[(budget_year, units)]
                    for year, units in df_budget.columns
                    if year < budget_year
                }
            )
            df_budget.update(df_exceedance_budget)
        # pd.where operation requires DataFrames to be aligned
        df_aligned = df_subject.where(df_subject <= df_budget)
        # Drop the embedded units from the multi-index and find the first (meaning furthest-out) date of alignment
        df_aligned = df_aligned.droplevel(1, axis=1).apply(lambda x: x.first_valid_index(), axis=1)

        if len(missing_subjects):
            df_aligned = pd.concat(
                [
                    df_aligned,
                    pd.Series(data=[pd.NA] * len(missing_subjects), index=missing_subjects),
                ]
            )
        df_exceedance = df_aligned.map(lambda x: base_year if ITR.isna(x) else pd.NA if x >= target_year else x).astype(
            "Int64"
        )
        return df_exceedance
